# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Contacts list view - shown in the sidebar."""

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")
gi.require_version("Shumate", "1.0")
from gi.repository import Adw, Gio, GLib, GObject, Gtk, Shumate
from meshy.models import Contact, ContactType
from meshy.protocol import build_sync_next_message
from meshy.views import (
    _add_hover_swap,
    _contact_pick_factory_bind,
    _contact_pick_factory_setup,
    _contact_pick_factory_unbind,
    _ContactPickItem,
)


@Gtk.Template(resource_path="/page/codeberg/sesivany/Meshy/ui/contacts-view.ui")
class ContactsView(Gtk.Box):
    __gtype_name__ = "MeshyContactsView"

    _search_bar = Gtk.Template.Child("search_bar")
    _search_entry = Gtk.Template.Child("search_entry")
    _list_box = Gtk.Template.Child("list_box")
    _status_label = Gtk.Template.Child("status_label")
    _add_btn = Gtk.Template.Child("add_btn")

    # Filter constants
    FILTER_ALL = "all"
    FILTER_FAVORITES = "favorites"
    FILTER_USERS = "users"
    FILTER_REPEATERS = "repeaters"
    FILTER_ROOMS = "rooms"
    FILTER_SENSORS = "sensors"

    # Sort constants
    SORT_AZ = "az"
    SORT_HEARD = "heard"
    SORT_MESSAGES = "messages"

    def __init__(self, window, **kwargs):
        super().__init__(**kwargs)
        self._window = window
        self._contact_rows: dict[str, Adw.ActionRow] = {}
        self._unread_keys: set[str] = set()
        self._unread_counts: dict[str, int] = {}
        self._unread_dots: dict[str, Gtk.Label] = {}
        self._current_filter = self.FILTER_ALL
        self._current_sort = self.SORT_HEARD
        self._favorites_first = True
        self._login_dialog_open = False
        self._total_contacts = 0

        # Register actions
        action_group = Gio.SimpleActionGroup()
        filter_action = Gio.SimpleAction.new_stateful(
            "filter", GLib.VariantType.new("s"), GLib.Variant.new_string(self._current_filter)
        )
        filter_action.connect("activate", self._on_filter_changed)
        action_group.add_action(filter_action)
        sort_action = Gio.SimpleAction.new_stateful(
            "sort", GLib.VariantType.new("s"), GLib.Variant.new_string(self._current_sort)
        )
        sort_action.connect("activate", self._on_sort_changed)
        action_group.add_action(sort_action)
        fav_action = Gio.SimpleAction.new_stateful(
            "favorites-first", None, GLib.Variant.new_boolean(True)
        )
        fav_action.connect("change-state", self._on_favorites_first_toggled)
        action_group.add_action(fav_action)

        # Add contact menu (conditional QR item)
        add_menu = Gio.Menu()
        add_menu.append(_("Discover Contacts"), "contacts.discover")
        add_menu.append(_("Add Manually"), "contacts.add-manual")
        add_menu.append(_("Import from Clipboard"), "contacts.import-clipboard")
        import meshy as _meshy_pkg

        if _meshy_pkg.QR_SCANNER_ENABLED:
            add_menu.append(_("Scan QR Code"), "contacts.scan-qr")
        self._add_btn.set_menu_model(add_menu)

        add_action = Gio.SimpleAction.new("add-manual", None)
        add_action.connect("activate", self._on_add_manual)
        action_group.add_action(add_action)

        discover_action = Gio.SimpleAction.new("discover", None)
        discover_action.connect("activate", self._on_discover_contacts)
        action_group.add_action(discover_action)

        clipboard_action = Gio.SimpleAction.new("import-clipboard", None)
        clipboard_action.connect("activate", self._on_import_clipboard)
        action_group.add_action(clipboard_action)

        if _meshy_pkg.QR_SCANNER_ENABLED:
            qr_action = Gio.SimpleAction.new("scan-qr", None)
            qr_action.connect("activate", self._on_scan_qr)
            action_group.add_action(qr_action)

        self._action_group = action_group
        self.insert_action_group("contacts", action_group)
        self._search_btn = None
        self._search_binding = None

        # Signal connections
        self._search_bar.connect_entry(self._search_entry)
        self._search_bar.connect("notify::search-mode-enabled", self._on_search_mode_changed)
        self._search_entry.connect("search-changed", self._on_search_changed)
        self._list_box.set_filter_func(self._filter_func)
        self._list_box.connect("row-activated", self._on_row_selected)

        # Initial populate
        self.update_contacts(self._window.contacts)

    def set_search_button(self, btn):
        if self._search_binding:
            self._search_binding.unbind()
        self._search_btn = btn
        self._search_binding = btn.bind_property(
            "active",
            self._search_bar,
            "search-mode-enabled",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )

    def build_filter_menu(self):
        menu = Gio.Menu()
        for filter_id, label in (
            ("all", _("All")),
            ("favorites", _("Favourites")),
            ("users", _("Users")),
            ("repeaters", _("Repeaters")),
            ("rooms", _("Room Servers")),
            ("sensors", _("Sensors")),
        ):
            item = Gio.MenuItem.new(label, None)
            item.set_action_and_target_value("contacts.filter", GLib.Variant.new_string(filter_id))
            menu.append_item(item)
        return menu

    def build_sort_menu(self):
        menu = Gio.Menu()
        section1 = Gio.Menu()
        for sort_id, label in (
            ("az", _("A–Z")),
            ("heard", _("Heard Recently")),
            ("messages", _("Latest Messages")),
        ):
            item = Gio.MenuItem.new(label, None)
            item.set_action_and_target_value("contacts.sort", GLib.Variant.new_string(sort_id))
            section1.append_item(item)
        menu.append_section(None, section1)
        section2 = Gio.Menu()
        section2.append(_("Favorites First"), "contacts.favorites-first")
        menu.append_section(None, section2)
        return menu

    def _filter_func(self, row) -> bool:
        if not hasattr(row, "_contact"):
            return True
        contact = row._contact

        # Type filter
        f = self._current_filter
        if (
            f == self.FILTER_FAVORITES
            and not contact.is_favorite
            or f == self.FILTER_USERS
            and contact.type != ContactType.CHAT
            or f == self.FILTER_REPEATERS
            and contact.type != ContactType.REPEATER
            or f == self.FILTER_ROOMS
            and contact.type != ContactType.ROOM
            or f == self.FILTER_SENSORS
            and contact.type != ContactType.SENSOR
        ):
            return False

        # Search filter
        search = self._search_entry.get_text().lower()
        return not (search and search not in contact.name.lower())

    def _update_status_count(self):
        """Update the status label with visible/total contact count."""
        total = self._total_contacts
        if self._current_filter != self.FILTER_ALL or self._search_entry.get_text():
            # Count rows that pass the filter
            visible = sum(1 for row in self._contact_rows.values() if self._filter_func(row))
            self._status_label.set_label(
                _("{visible}/{total} contacts").format(visible=visible, total=total)
            )
        else:
            self._status_label.set_label(
                _("{total} contacts").format(total=total)
                if total != 1
                else _("{total} contact").format(total=total)
            )

    def _on_search_mode_changed(self, search_bar, pspec):
        if not search_bar.get_search_mode():
            self._search_entry.set_text("")

    def _on_search_changed(self, entry):
        self._list_box.invalidate_filter()
        GLib.idle_add(self._update_status_count)

    def _on_filter_changed(self, action, param):
        self._current_filter = param.get_string()
        action.set_state(param)
        self._list_box.invalidate_filter()
        GLib.idle_add(self._update_status_count)

    def _on_sort_changed(self, action, param):
        self._current_sort = param.get_string()
        action.set_state(param)
        self.update_contacts(self._window.contacts)

    def _on_favorites_first_toggled(self, action, value):
        action.set_state(value)
        self._favorites_first = value.get_boolean()
        self.update_contacts(self._window.contacts)

    def _on_discover_contacts(self, *args):
        discovered = [
            c
            for c in self._window.discovered_contacts
            if not self._window.is_contact_known(c.public_key_hex)
        ]
        if not discovered:
            dialog = Adw.AlertDialog(
                heading=_("Discover Contacts"),
                body=_(
                    "No new contacts have been heard yet. "
                    "Leave the app running and contacts will appear "
                    "as their adverts are received."
                ),
            )
            dialog.add_response("ok", _("OK"))
            dialog.present(self._window)
            return

        builder = Gtk.Builder.new_from_resource(
            "/page/codeberg/sesivany/Meshy/ui/discover-dialog.ui"
        )
        dialog = builder.get_object("discover_dialog")
        search_btn = builder.get_object("discover_search_btn")
        filter_btn = builder.get_object("discover_filter_btn")
        sort_btn = builder.get_object("discover_sort_btn")
        search_entry = builder.get_object("discover_search_entry")
        list_view = builder.get_object("discover_list_view")
        status_label = builder.get_object("discover_status_label")

        # Search toggle
        search_btn.bind_property(
            "active", search_entry, "visible", GObject.BindingFlags.SYNC_CREATE
        )
        search_entry.connect("stop-search", lambda _e: search_btn.set_active(False))

        # Filter menu
        filter_menu = Gio.Menu()
        for fid, label in (
            ("all", _("All")),
            ("users", _("Users")),
            ("repeaters", _("Repeaters")),
            ("rooms", _("Room Servers")),
            ("sensors", _("Sensors")),
        ):
            item = Gio.MenuItem.new(label, f"discover.filter::{fid}")
            filter_menu.append_item(item)
        filter_btn.set_menu_model(filter_menu)

        # Sort menu
        dev = self._window.device_info
        has_self_loc = (
            dev
            and dev.latitude is not None
            and dev.longitude is not None
            and (abs(dev.latitude) > 1e-6 or abs(dev.longitude) > 1e-6)
        )
        sort_menu = Gio.Menu()
        for sid, label in (
            ("name", _("Name")),
            ("last-seen", _("Last Seen")),
            ("arrival", _("Arrival Time")),
        ):
            sort_menu.append_item(Gio.MenuItem.new(label, f"discover.sort::{sid}"))
        if has_self_loc:
            sort_menu.append_item(Gio.MenuItem.new(_("Distance"), "discover.sort::distance"))
        sort_btn.set_menu_model(sort_menu)

        action_group = Gio.SimpleActionGroup()
        filter_action = Gio.SimpleAction.new_stateful(
            "filter", GLib.VariantType.new("s"), GLib.Variant.new_string("all")
        )
        sort_action = Gio.SimpleAction.new_stateful(
            "sort", GLib.VariantType.new("s"), GLib.Variant.new_string("name")
        )
        action_group.add_action(filter_action)
        action_group.add_action(sort_action)
        dialog.insert_action_group("discover", action_group)

        # Model — prevent GC by storing refs on the dialog
        store = Gio.ListStore(item_type=_ContactPickItem)
        for c in sorted(discovered, key=lambda c: c.name.lower()):
            store.append(_ContactPickItem(c))

        # Filter
        custom_filter = Gtk.CustomFilter.new(
            self._discover_filter_func, (search_entry, filter_action)
        )

        def _on_filter_or_search_changed(*_args):
            custom_filter.changed(Gtk.FilterChange.DIFFERENT)
            self._update_discover_status(status_label, filter_model, store)

        def _on_filter_activated(_action, param):
            filter_action.set_state(param)
            _on_filter_or_search_changed()

        search_entry.connect("search-changed", _on_filter_or_search_changed)
        filter_action.connect("activate", _on_filter_activated)

        filter_model = Gtk.FilterListModel(model=store, filter=custom_filter)

        # Sorter
        self_lat = dev.latitude if dev else None
        self_lon = dev.longitude if dev else None
        custom_sorter = Gtk.CustomSorter.new(
            self._discover_sort_func, (sort_action, self_lat, self_lon)
        )

        def _on_sort_activated(_action, param):
            sort_action.set_state(param)
            custom_sorter.changed(Gtk.SorterChange.DIFFERENT)

        sort_action.connect("activate", _on_sort_activated)

        sort_model = Gtk.SortListModel(model=filter_model, sorter=custom_sorter)

        selection = Gtk.NoSelection(model=sort_model)
        list_view.set_model(selection)

        # Factory
        def _on_add(contact):
            item = None
            for i in range(store.get_n_items()):
                it = store.get_item(i)
                if it.contact.public_key_hex == contact.public_key_hex:
                    item = it
                    break
            if item:
                self._add_discovered_lv(item, store, dialog, status_label, filter_model)

        def _subtitle(contact):
            subtitle = contact.type_label
            if has_self_loc and contact.has_location:
                from meshy.utils import haversine_distance

                dist = haversine_distance(self_lat, self_lon, contact.latitude, contact.longitude)
                if dist < 1.0:
                    subtitle += f" · {dist * 1000:.0f} m"
                else:
                    subtitle += f" · {dist:.1f} km"
            return subtitle

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", lambda f, li: _contact_pick_factory_setup(f, li, _("Add")))
        factory.connect(
            "bind",
            lambda f, li: _contact_pick_factory_bind(f, li, _on_add, subtitle_func=_subtitle),
        )
        factory.connect("unbind", _contact_pick_factory_unbind)
        list_view.set_factory(factory)

        def _on_activate(lv, pos):
            item = sort_model.get_item(pos)
            if item:
                self._add_discovered_lv(item, store, dialog, status_label, filter_model)

        list_view.connect("activate", _on_activate)

        # prevent Python GC from collecting these while the dialog is alive
        dialog._refs = (
            store,
            custom_filter,
            filter_model,
            custom_sorter,
            sort_model,
            selection,
            factory,
            action_group,
        )

        self._update_discover_status(status_label, filter_model, store)
        dialog.present(self._window)

    def _discover_filter_func(self, item, user_data):
        search_entry, filter_action = user_data
        contact = item.contact
        search = search_entry.get_text().lower()
        if search and search not in contact.name.lower():
            return False
        f = filter_action.get_state().get_string()
        return not (
            f == "users"
            and contact.type != ContactType.CHAT
            or f == "repeaters"
            and contact.type != ContactType.REPEATER
            or f == "rooms"
            and contact.type != ContactType.ROOM
            or f == "sensors"
            and contact.type != ContactType.SENSOR
        )

    @staticmethod
    def _discover_sort_func(a, b, user_data=None):
        sort_action, self_lat, self_lon = user_data
        mode = sort_action.get_state().get_string()

        if mode == "last-seen":
            ta = a.contact.last_seen
            tb = b.contact.last_seen
            if ta is not None and tb is not None:
                return (tb > ta) - (tb < ta)
            if ta is not None:
                return -1
            if tb is not None:
                return 1
            na = a.contact.name.lower()
            nb = b.contact.name.lower()
            return (na > nb) - (na < nb)

        if mode == "arrival":
            ta = a.contact.received_at
            tb = b.contact.received_at
            if ta is not None and tb is not None:
                return (tb > ta) - (tb < ta)
            if ta is not None:
                return -1
            if tb is not None:
                return 1
            na = a.contact.name.lower()
            nb = b.contact.name.lower()
            return (na > nb) - (na < nb)

        if mode == "distance" and self_lat is not None and self_lon is not None:
            from meshy.utils import haversine_distance

            da = (
                haversine_distance(self_lat, self_lon, a.contact.latitude, a.contact.longitude)
                if a.contact.has_location
                else float("inf")
            )
            db = (
                haversine_distance(self_lat, self_lon, b.contact.latitude, b.contact.longitude)
                if b.contact.has_location
                else float("inf")
            )
            if da != db:
                return (da > db) - (da < db)

        na = a.contact.name.lower()
        nb = b.contact.name.lower()
        return (na > nb) - (na < nb)

    def _add_discovered_lv(self, item, store, dialog, status_label, filter_model):
        self._window.add_discovered_contact(item.contact)
        found, pos = store.find(item)
        if found:
            store.remove(pos)
        self._update_discover_status(status_label, filter_model, store)
        if store.get_n_items() == 0:
            dialog.close()

    @staticmethod
    def _update_discover_status(label, filter_model, store):
        total = store.get_n_items()
        visible = filter_model.get_n_items()
        if total == 0:
            label.set_label(_("No contacts"))
        elif visible != total:
            label.set_label(_("{visible}/{total} discovered").format(visible=visible, total=total))
        else:
            label.set_label(_("{total} discovered").format(total=total))

    def _on_add_manual(self, *args):
        dialog = Adw.AlertDialog(
            heading=_("Add Contact"),
            body=_("Enter the contact details."),
        )
        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)

        type_row = Adw.ComboRow(title=_("Contact Type"))
        type_list = Gtk.StringList()
        for label in [_("Chat"), _("Repeater"), _("Room Server"), _("Sensor")]:
            type_list.append(label)
        type_row.set_model(type_list)
        type_row.set_selected(0)

        name_row = Adw.EntryRow(title=_("Name"))
        key_row = Adw.EntryRow(title=_("Public Key (64 hex characters)"))

        fields_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        fields_list.add_css_class("boxed-list")
        fields_list.append(type_row)
        fields_list.append(name_row)
        fields_list.append(key_row)
        content.append(fields_list)

        dialog.set_extra_child(content)
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("add", _("Add"))
        dialog.set_response_appearance("add", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "add":
                try:
                    name = name_row.get_text().strip()
                    key_text = key_row.get_text().strip()
                    pub_key = bytes.fromhex(key_text)
                    if len(pub_key) != 32 or not name:
                        return
                    if self._window.is_contact_known(key_text):
                        self._show_import_error(_("This contact is already in your list."))
                        return
                    contact_type = (
                        type_row.get_selected() + 1
                    )  # CHAT=1, REPEATER=2, ROOM=3, SENSOR=4
                    self._window.add_contact(pub_key, contact_type, name)
                except ValueError:
                    pass

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _on_import_clipboard(self, *args):
        clipboard = self._window.get_clipboard()
        clipboard.read_text_async(None, self._on_clipboard_read)

    def _on_clipboard_read(self, clipboard, result):
        try:
            text = clipboard.read_text_finish(result)
        except Exception:
            text = None

        if not text or not text.strip().startswith("meshcore://"):
            dialog = Adw.AlertDialog(
                heading=_("Import Failed"),
                body=_("No meshcore:// link found in clipboard."),
            )
            dialog.add_response("ok", _("OK"))
            dialog.present(self._window)
            return

        app = self._window.get_application()
        app._handle_meshcore_uri(text.strip())

    def _on_scan_qr(self, *args):
        from meshy.camera import create_qr_scanner

        self._qr_scanner = create_qr_scanner(self._window, self._on_qr_result)
        self._qr_scanner.start()

    def _on_qr_result(self, data):
        """Handle a scanned QR code - meshcore:// URL or raw public key hex."""
        from urllib.parse import parse_qs, urlparse

        # meshcore://contact/add?name=...&public_key=...
        if data.lower().startswith("meshcore://"):
            parsed = urlparse(data)
            params = parse_qs(parsed.query)

            full_path = f"{parsed.netloc}{parsed.path}".rstrip("/")
            if full_path == "channel/add" and "secret" in params:
                # meshcore://channel/add?name=...&secret=...
                ch_name = params.get("name", [""])[0]
                secret_hex = params["secret"][0].strip()
                try:
                    secret = bytes.fromhex(secret_hex)
                except ValueError:
                    self._show_import_error(_("Invalid channel secret in QR code."))
                    return
                if len(secret) != 16:
                    self._show_import_error(
                        _("Channel secret must be 16 bytes, got {}.").format(len(secret))
                    )
                    return
                self._window.add_channel_from_qr(ch_name, secret)
                return

            if "public_key" in params:
                pub_key_hex = params["public_key"][0].strip()
                name = params.get("name", [""])[0]
                # Extract contact type: 1=Companion, 2=Repeater, 3=RoomServer, 4=Sensor
                try:
                    contact_type = int(params.get("type", ["1"])[0])
                except ValueError:
                    contact_type = 1
                try:
                    pub_key = bytes.fromhex(pub_key_hex)
                except ValueError:
                    self._show_import_error(_("Invalid public key in QR code."))
                    return

                if len(pub_key) != 32:
                    self._show_import_error(
                        _("Public key must be 32 bytes, got {}.").format(len(pub_key))
                    )
                    return

                if self._window.is_contact_known(pub_key_hex):
                    self._show_import_error(_("This contact is already in your list."))
                    return

                if name:
                    self._window.add_contact(pub_key, contact_type, name)
                else:
                    self._show_qr_contact_dialog(pub_key, pub_key_hex)
                return

            # Try as raw hex contact frame (old format)
            hex_data = data[len("meshcore://") :].strip()
            if "/" not in hex_data and "?" not in hex_data:
                try:
                    contact_frame = bytes.fromhex(hex_data)
                    self._window.import_contact(contact_frame)
                    return
                except ValueError:
                    pass

        # Try as raw hex public key (64 hex chars = 32 bytes)
        cleaned = data.strip().replace(" ", "").replace(":", "")
        try:
            pub_key = bytes.fromhex(cleaned)
            if len(pub_key) == 32:
                if self._window.is_contact_known(cleaned):
                    self._show_import_error(_("This contact is already in your list."))
                    return
                self._show_qr_contact_dialog(pub_key, cleaned)
                return
        except ValueError:
            pass

        self._show_import_error(_("Could not parse QR code data:\n{}").format(data[:120]))

    def _show_import_error(self, message):
        dialog = Adw.AlertDialog(heading=_("Import Failed"), body=message)
        dialog.add_response("ok", _("OK"))
        dialog.present(self._window)

    def _show_qr_contact_dialog(self, pub_key, hex_str):
        """Show dialog to complete contact details from a scanned public key."""
        dialog = Adw.AlertDialog(
            heading=_("QR Code Scanned"),
            body=_("Public key: {}...{}").format(hex_str[:16], hex_str[-16:]),
        )
        content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=8)

        type_row = Adw.ComboRow(title=_("Contact Type"))
        type_list = Gtk.StringList()
        for label in [_("Chat"), _("Repeater"), _("Room Server"), _("Sensor")]:
            type_list.append(label)
        type_row.set_model(type_list)
        type_row.set_selected(0)

        name_row = Adw.EntryRow(title=_("Name"))

        fields_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        fields_list.add_css_class("boxed-list")
        fields_list.append(type_row)
        fields_list.append(name_row)
        content.append(fields_list)

        dialog.set_extra_child(content)
        dialog.add_response("cancel", _("Cancel"))
        dialog.add_response("add", _("Add Contact"))
        dialog.set_response_appearance("add", Adw.ResponseAppearance.SUGGESTED)

        def _on_response(d, response):
            if response == "add":
                name = name_row.get_text().strip()
                if not name:
                    return
                contact_type = type_row.get_selected() + 1
                self._window.add_contact(pub_key, contact_type, name)

        dialog.connect("response", _on_response)
        dialog.present(self._window)

    def _on_row_selected(self, listbox, row):
        import logging

        log = logging.getLogger(__name__)
        if self._login_dialog_open:
            log.debug("Login dialog already open, ignoring row selection")
            return
        if row and hasattr(row, "_contact"):
            contact = row._contact
            log.info(
                f"Contact selected: {contact.name} (type={contact.type}, logged_in={self._window.is_room_logged_in(contact.public_key_hex)})"
            )
            if contact.type == ContactType.REPEATER:
                if self._window.is_room_logged_in(contact.public_key_hex):
                    self._window.show_repeater(
                        contact, is_admin=self._window.is_room_admin(contact.public_key_hex)
                    )
                else:
                    self._show_room_login(contact, repeater=True)
            elif contact.type == ContactType.ROOM and not self._window.is_room_logged_in(
                contact.public_key_hex
            ):
                log.info("Room not logged in, showing login dialog")
                self._show_room_login(contact)
            else:
                self._window.show_chat(contact)

    def _show_room_login(self, contact: Contact, management: bool = False, repeater: bool = False):
        """Show a login dialog for a room/repeater contact."""
        import logging

        log = logging.getLogger(__name__)
        log.info(
            f"Showing login dialog for {contact.name} (type={contact.type}, pub_key={contact.public_key_hex[:16]}...)"
        )
        self._login_dialog_open = True
        dialog = Adw.Dialog()
        dialog.set_title(_("Login to {}").format(contact.name))
        dialog.set_content_width(360)
        dialog.set_content_height(280)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=12,
            margin_start=16,
            margin_end=16,
            margin_top=12,
            margin_bottom=12,
        )

        password_entry = Adw.PasswordEntryRow(title=_("Password"))
        save_switch = Adw.SwitchRow(title=_("Save password"))

        login_list = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        login_list.add_css_class("boxed-list")
        login_list.append(password_entry)
        login_list.append(save_switch)
        content.append(login_list)

        # Pre-fill saved password
        if self._window.storage:
            saved = self._window.storage.get_room_password(contact.public_key_hex)
            if saved:
                password_entry.set_text(saved)
                save_switch.set_active(True)

        status_box = Gtk.Box(spacing=8, halign=Gtk.Align.CENTER, margin_top=4)
        spinner = Gtk.Spinner()
        status_box.append(spinner)
        status_label = Gtk.Label(label="")
        status_label.add_css_class("dim-label")
        status_box.append(status_label)
        content.append(status_box)

        login_btn = Gtk.Button(label=_("Login"))
        login_btn.add_css_class("suggested-action")
        login_btn.set_halign(Gtk.Align.CENTER)
        content.append(login_btn)

        toolbar_view.set_content(content)
        dialog.set_child(toolbar_view)

        timeout_id = [None]

        def on_login_result(success, login_timestamp=None, is_admin=True):
            spinner.set_spinning(False)
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None
            if success:
                status_label.set_label(_("Login successful"))
                # Save or remove password
                if self._window.storage:
                    if save_switch.get_active():
                        self._window.storage.save_room_password(
                            contact.public_key_hex, password_entry.get_text()
                        )
                    else:
                        self._window.storage.remove_room_password(contact.public_key_hex)
                self._window.mark_room_logged_in(contact.public_key_hex, is_admin)
                # Immediately sync to fetch pending room messages
                self._window.send_frame(build_sync_next_message())
                dialog.close()
                if repeater:
                    self._window.show_repeater(contact, login_timestamp, is_admin)
                elif management:
                    self._show_room_management(contact)
                else:
                    self._window.show_chat(contact)
            else:
                status_label.set_label(_("Login failed — wrong password"))
                status_label.remove_css_class("dim-label")
                status_label.add_css_class("error")
                login_btn.set_sensitive(True)

        def on_timeout():
            spinner.set_spinning(False)
            status_label.set_label(_("Login timed out"))
            self._window.stop_login()
            login_btn.set_sensitive(True)
            timeout_id[0] = None
            return False

        def on_retry():
            status_label.set_label(_("Retrying via flood..."))

        def on_login_clicked(*args):
            log.info(f"Login button clicked for {contact.name}")
            password = password_entry.get_text()
            log.info(f"Password length: {len(password)}")
            spinner.set_spinning(True)
            status_label.set_label(_("Logging in..."))
            status_label.remove_css_class("error")
            status_label.add_css_class("dim-label")
            login_btn.set_sensitive(False)
            timeout_id[0] = GLib.timeout_add(15000, on_timeout)
            log.info(f"Calling send_login with pub_key={contact.public_key.hex()[:16]}...")
            self._window.send_login(contact.public_key, password, on_login_result, on_retry)

        login_btn.connect("clicked", on_login_clicked)
        password_entry.connect("apply", on_login_clicked)

        def on_dialog_closed(d):
            self._login_dialog_open = False
            self._window.stop_login()
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])

        dialog.connect("closed", on_dialog_closed)
        dialog.present(self._window)

    def _sort_key(self, c: Contact):
        """Compute sort key for a contact based on current sort setting."""
        fav = self._favorites_first
        if self._current_sort == self.SORT_AZ:
            return (not c.is_favorite if fav else False, c.name.lower())
        elif self._current_sort == self.SORT_MESSAGES:
            return (
                not c.is_favorite if fav else False,
                -(c.last_message_at.timestamp() if c.last_message_at else 0),
            )
        else:  # SORT_HEARD
            return (
                not c.is_favorite if fav else False,
                -(c.last_seen.timestamp() if c.last_seen else 0),
            )

    def update_contacts(self, contacts: list[Contact]):
        sorted_contacts = sorted(contacts, key=self._sort_key)
        new_keys = [c.public_key_hex for c in sorted_contacts]
        old_keys = list(self._contact_rows.keys())

        # Fast path: same contacts in same order — update in-place
        if new_keys == old_keys:
            for contact in sorted_contacts:
                key = contact.public_key_hex
                row = self._contact_rows.get(key)
                if row:
                    row.set_title(GLib.markup_escape_text(contact.name))
                    row.set_subtitle(contact.path_label)
                    row._contact = contact
            self._total_contacts = len(contacts)
            self._update_status_count()
            return

        # Full rebuild: contact set or order changed
        self._full_rebuild(sorted_contacts)

    def _full_rebuild(self, sorted_contacts: list[Contact]):
        """Remove all rows and recreate from scratch."""
        selected_row = self._list_box.get_selected_row()
        selected_key = (
            selected_row._contact.public_key_hex
            if selected_row and hasattr(selected_row, "_contact")
            else None
        )

        child = self._list_box.get_first_child()
        while child:
            next_child = child.get_next_sibling()
            self._list_box.remove(child)
            child = next_child
        self._contact_rows.clear()

        new_selected = None
        for contact in sorted_contacts:
            row = self._create_contact_row(contact)
            self._list_box.append(row)
            self._contact_rows[contact.public_key_hex] = row
            if contact.public_key_hex == selected_key:
                new_selected = row

        if new_selected:
            self._list_box.select_row(new_selected)

        self._total_contacts = len(sorted_contacts)
        self._update_status_count()

    def _create_contact_row(self, contact: Contact) -> Gtk.ListBoxRow:
        row = Adw.ActionRow(
            title=GLib.markup_escape_text(contact.name),
            subtitle=contact.path_label,
        )
        row._contact = contact

        # Type icon with favorite star
        icon_name = {
            ContactType.CHAT: "avatar-default-symbolic",
            ContactType.REPEATER: "network-server-symbolic",
            ContactType.ROOM: "user-home-symbolic",
            ContactType.SENSOR: "weather-clear-symbolic",
        }.get(contact.type, "avatar-default-symbolic")
        prefix_box = Gtk.Box(spacing=2)
        star = Gtk.Image.new_from_icon_name("starred-symbolic")
        star.set_pixel_size(10)
        if not contact.is_favorite:
            star.set_opacity(0)
        prefix_box.append(star)
        prefix_box.append(Gtk.Image.new_from_icon_name(icon_name))
        row.add_prefix(prefix_box)

        # Unread badge — visible by default when count > 0, hidden on hover
        count = self._unread_counts.get(contact.public_key_hex, 0)
        unread_badge = Gtk.Label(label=str(count) if count else "")
        unread_badge.add_css_class("unread-badge")
        unread_badge.set_valign(Gtk.Align.CENTER)
        unread_badge.set_visible(count > 0)
        row.add_suffix(unread_badge)
        self._unread_dots[contact.public_key_hex] = unread_badge

        # "More" button + hover behavior + long press
        info_btn = Gtk.Button(icon_name="view-more-symbolic", valign=Gtk.Align.CENTER)
        info_btn.add_css_class("flat")
        info_btn.set_visible(False)
        on_detail = lambda *_, r=row: self._show_contact_detail(r._contact)
        info_btn.connect("clicked", on_detail)
        row.add_suffix(info_btn)
        _add_hover_swap(
            row,
            info_btn,
            unread_badge,
            lambda: self._unread_counts.get(contact.public_key_hex, 0),
            on_detail,
        )

        row.set_activatable(True)
        return row

    def mark_unread(self, public_key_hex: str, unread: bool):
        """Show or hide the unread badge for a contact."""
        if unread:
            self._unread_keys.add(public_key_hex)
            self._unread_counts[public_key_hex] = self._unread_counts.get(public_key_hex, 0) + 1
        else:
            self._unread_keys.discard(public_key_hex)
            self._unread_counts[public_key_hex] = 0
        if public_key_hex in self._unread_dots:
            badge = self._unread_dots[public_key_hex]
            count = self._unread_counts.get(public_key_hex, 0)
            badge.set_label(str(count) if count else "")
            badge.set_visible(count > 0)
        self._window.update_nav_unread_contacts()

    def _get_visible_rows(self):
        """Return visible rows in display order."""
        rows = []
        row = self._list_box.get_first_child()
        while row:
            if row.get_visible() and hasattr(row, "_contact"):
                rows.append(row)
            row = row.get_next_sibling()
        return rows

    def navigate(self, direction: int, unread_only: bool = False):
        """Move selection up (-1) or down (+1). If unread_only, skip to next unread."""
        rows = self._get_visible_rows()
        if not rows:
            return
        if unread_only:
            rows = [r for r in rows if r._contact.public_key_hex in self._unread_keys]
            if not rows:
                return
        selected = self._list_box.get_selected_row()
        if selected and selected in rows:
            idx = rows.index(selected)
            idx = (idx + direction) % len(rows)
        else:
            idx = 0 if direction > 0 else len(rows) - 1
        self._list_box.select_row(rows[idx])
        rows[idx].grab_focus()

    def _on_path_discovery(self, contact: Contact, btn: Gtk.Button):
        """Send a path discovery request for a contact."""
        btn.set_sensitive(False)
        self._discover_path_label.set_label(_("Searching..."))

        timeout_id = [None]

        def _on_response(data):
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None
            self._window.stop_path_discovery()
            btn.set_sensitive(True)
            # Parse the PATH_DISCOVERY_RESPONSE (0x8D)
            # Format: [code][reserved][prefix×6][out_path_len][out_path...][in_path_len][in_path...]
            try:
                i = 2  # skip code + reserved
                i += 6  # skip prefix
                opl_byte = data[i]
                i += 1
                opl_hlen = ((opl_byte & 0xC0) >> 6) + 1
                opl = opl_byte & 0x3F
                out_hops = []
                for _j in range(opl):
                    out_hops.append(data[i : i + opl_hlen].hex().upper())
                    i += opl_hlen
                ipl_byte = data[i]
                i += 1
                ipl_hlen = ((ipl_byte & 0xC0) >> 6) + 1
                ipl = ipl_byte & 0x3F
                in_hops = []
                for _j in range(ipl):
                    in_hops.append(data[i : i + ipl_hlen].hex().upper())
                    i += ipl_hlen

                if opl == 0 and ipl == 0:
                    self._discover_path_label.set_label(_("Direct"))
                else:
                    parts = []
                    if out_hops:
                        parts.append(",".join(out_hops))
                    self._discover_path_label.set_label(
                        " → ".join(parts) if parts else f"{opl} hops"
                    )

                    # Apply discovered outbound path to the contact
                    if out_hops:
                        path_bytes = b"".join(bytes.fromhex(h) for h in out_hops)
                        hash_mode = opl_hlen - 1
                        path_len_byte = (hash_mode << 6) | (opl & 0x3F)
                        self._window.set_contact_path(contact, path_len_byte, path_bytes)
                        self._window.show_toast(f"Path set: {','.join(out_hops)}")
            except (IndexError, ValueError) as e:
                import logging

                logging.getLogger(__name__).warning(f"Path discovery parse error: {e}")
                self._discover_path_label.set_label(_("Path found"))
            self._window.refresh_contacts()

        def _on_timeout():
            self._window.stop_path_discovery()
            btn.set_sensitive(True)
            self._discover_path_label.set_label(_("No response"))
            timeout_id[0] = None
            return False

        timeout_id[0] = GLib.timeout_add(30000, _on_timeout)
        self._window.send_path_discovery(contact.public_key, _on_response)

    def _on_get_advert_path(self, contact: Contact):
        """Get and display the inbound advertisement path for a contact."""
        self._advert_path_label.set_label(_("Loading..."))

        # Set a timeout in case the response never comes (error consumed elsewhere)
        timeout_id = [None]

        def _on_response(result):
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None
            if not result:
                self._advert_path_label.set_label(_("No path cached"))
                return
            hops = result.get("hops", [])
            path_len = result.get("path_length", -1)
            if path_len < 0 or not hops:
                self._advert_path_label.set_label(_("Flood"))
            elif path_len == 0:
                self._advert_path_label.set_label(_("Direct"))
            else:
                self._advert_path_label.set_label(" → ".join(hops))

        def _on_timeout():
            self._window.set_advert_path_callback(None)
            self._advert_path_label.set_label(_("Not supported"))
            timeout_id[0] = None
            return False

        timeout_id[0] = GLib.timeout_add(3000, _on_timeout)
        self._window.get_advert_path(contact.public_key, _on_response)

    def _ping_contact(self, contact: Contact, btn: Gtk.Button):
        """Ping a repeater/room via trace path to get RTT and SNR."""
        # Trace protocol only supports 1B and 2B hashes (power-of-2 encoding)
        if self._window.device_info.path_hash_mode > 1:
            self._window.show_toast(_("Ping is not supported with 3-byte path hashes"))
            return

        import time

        btn.set_sensitive(False)
        start = time.monotonic()
        timeout_id = [None]

        hash_width = self._window.device_info.path_hash_mode + 1
        hop_hash = contact.public_key[:hash_width]

        def on_trace(result):
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None
            rtt = (time.monotonic() - start) * 1000
            snr_values = result.get("snr_values", [])
            parts = [f"{rtt:.0f} ms"]
            if len(snr_values) >= 1:
                parts.append(f"SNR out: {snr_values[0]:+.1f} dB")
            if len(snr_values) >= 2:
                parts.append(f"SNR back: {snr_values[-1]:+.1f} dB")
            msg = _("Ping {}: {}").format(contact.name, ", ".join(parts))
            self._window.show_toast(msg, timeout=5)
            btn.set_sensitive(True)

        def on_device_timeout(timeout_ms):
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
            timeout_id[0] = GLib.timeout_add(timeout_ms, on_timeout)

        def on_timeout():
            self._window.stop_trace_path()
            self._window.show_toast(_("Ping {}: timed out").format(contact.name), timeout=5)
            btn.set_sensitive(True)
            timeout_id[0] = None
            return False

        timeout_id[0] = GLib.timeout_add(15000, on_timeout)
        trace_flags = self._window.device_info.path_hash_mode
        self._window.send_trace_path(hop_hash, on_trace, on_device_timeout, flags=trace_flags)

    def _request_telemetry(self, contact: Contact, row: Adw.ActionRow):
        """Request and display telemetry from a contact."""
        from meshy.protocol import parse_cayenne_lpp

        row.set_subtitle(_("Requesting..."))

        timeout_id = [None]

        def on_timeout():
            row.set_subtitle(_("No response (timed out)"))
            timeout_id[0] = None
            return False

        def on_response(lpp_data):
            if timeout_id[0]:
                GLib.source_remove(timeout_id[0])
                timeout_id[0] = None
            if not lpp_data:
                row.set_subtitle(_("No telemetry data"))
                return
            entries = parse_cayenne_lpp(lpp_data)
            if not entries:
                row.set_subtitle(_("No telemetry data"))
                return
            self._show_telemetry_dialog(contact, entries)
            row.set_subtitle(_("Telemetry received"))

        timeout_id[0] = GLib.timeout_add(15000, on_timeout)
        self._window.get_telemetry(contact.public_key, on_response, contact_type=contact.type)

    @staticmethod
    def _battery_percent(volts: float) -> int:
        from meshy.models import battery_percent_from_mv

        return battery_percent_from_mv(int(volts * 1000))

    def _show_telemetry_dialog(self, contact: Contact, entries: list[dict]):
        """Show a dialog with parsed telemetry data and optional map."""
        from meshy.protocol import LPP_GPS, LPP_VOLTAGE

        # Check for GPS data
        gps_entry = next((e for e in entries if e["type"] == LPP_GPS), None)

        dialog = Adw.Dialog()
        dialog.set_title(_("Telemetry — {}").format(contact.name))
        dialog.set_content_width(420)
        dialog.set_content_height(550 if gps_entry else 400)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        scroll = Gtk.ScrolledWindow(vexpand=True)
        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=12,
            margin_start=12,
            margin_end=12,
            margin_top=12,
            margin_bottom=12,
        )

        channels = {}
        for entry in entries:
            channels.setdefault(entry["channel"], []).append(entry)

        for ch in sorted(channels):
            group = Adw.PreferencesGroup(title=_("Channel {}").format(ch))

            for entry in channels[ch]:
                val = entry["value"]
                if isinstance(val, dict):
                    alt_str = f", alt {val['alt']:.0f} m" if abs(val["alt"]) > 0.01 else ""
                    subtitle = f"{val['lat']:.4f}, {val['lon']:.4f}{alt_str}"
                elif entry["type"] == LPP_VOLTAGE and ch == 1:
                    pct = self._battery_percent(val)
                    subtitle = f"{pct}% ({val:.2f} V)"
                elif isinstance(val, float):
                    subtitle = f"{val:.2f} {entry['unit']}"
                else:
                    subtitle = f"{val} {entry['unit']}"

                title = _("Battery") if entry["type"] == LPP_VOLTAGE and ch == 1 else entry["name"]
                row = Adw.ActionRow(title=title, subtitle=subtitle)
                group.add(row)

            content.append(group)

        # Show map if GPS data is present
        if gps_entry:
            try:
                from meshy.views import create_shumate_map, deregister_map_widget

                lat = gps_entry["value"]["lat"]
                lon = gps_entry["value"]["lon"]

                from meshy.views.map_view import _MARKER_COLORS, _MARKER_ICONS, _SELF_COLOR, MapView

                map_widget, marker_layer, viewport = create_shumate_map(
                    zoom_level=14, lat=lat, lon=lon
                )
                map_widget.set_vexpand(False)
                map_widget.set_size_request(-1, 250)

                r, g, b = _MARKER_COLORS.get(contact.type, (0.5, 0.5, 0.5))
                icon_name = _MARKER_ICONS.get(contact.type, "avatar-default-symbolic")
                marker = MapView.create_marker(r, g, b, icon_name, contact.name, lat, lon)
                marker_layer.add_marker(marker)

                # Show our companion's position too
                lats, lons = [lat], [lon]
                si = self._window.self_info
                self_lat = si.get("adv_lat", 0.0)
                self_lon = si.get("adv_lon", 0.0)
                if abs(self_lat) > 1e-6 or abs(self_lon) > 1e-6:
                    self_name = si.get("name", "") or _("My Device")
                    self_marker = MapView.create_marker(
                        *_SELF_COLOR, "avatar-default-symbolic", self_name, self_lat, self_lon
                    )
                    marker_layer.add_marker(self_marker)
                    lats.append(self_lat)
                    lons.append(self_lon)

                # Fit viewport to show all markers
                from meshy.views.map_view import MapView

                MapView.fit_viewport(viewport, lats, lons)

                map_frame = Gtk.Frame()
                map_frame.set_child(map_widget)
                content.append(map_frame)

            except Exception as e:
                import logging

                log = logging.getLogger(__name__)
                log.error(f"Failed to create map widget for telemetry display: {e}", exc_info=True)
                # Show coordinates as text instead
                lat = gps_entry["value"]["lat"]
                lon = gps_entry["value"]["lon"]
                coords_label = Gtk.Label()
                # Avoid emoji on macOS due to CoreText rendering crashes (Bus error 10)
                from meshy.platforms import is_macos

                if is_macos():
                    coords_label.set_markup(
                        f"<b>GPS Coordinates:</b>\nLat: {lat:.6f}, Lon: {lon:.6f}"
                    )
                else:
                    coords_label.set_markup(f"<b>GPS Coordinates:</b>\n📍 {lat:.6f}, {lon:.6f}")
                coords_label.set_margin_top(12)
                coords_label.set_margin_bottom(12)
                content.append(coords_label)

        scroll.set_child(content)
        toolbar_view.set_content(scroll)
        dialog.set_child(toolbar_view)
        if gps_entry:
            dialog.connect("closed", lambda _d: deregister_map_widget(map_widget))
        dialog.present(self._window)

    def _show_room_management(self, contact: Contact):
        """Show a CLI management dialog for a room/repeater."""
        dialog = Adw.Dialog()
        dialog.set_title(_("{} — Management").format(contact.name))
        dialog.set_content_width(500)
        dialog.set_content_height(500)

        toolbar_view = Adw.ToolbarView()
        header = Adw.HeaderBar()
        toolbar_view.add_top_bar(header)

        content = Gtk.Box(
            orientation=Gtk.Orientation.VERTICAL,
            spacing=4,
            margin_start=8,
            margin_end=8,
            margin_bottom=8,
        )

        # Output area
        output_scroll = Gtk.ScrolledWindow(vexpand=True)
        output_view = Gtk.TextView(editable=False, wrap_mode=Gtk.WrapMode.WORD_CHAR)
        output_view.set_monospace(True)
        output_buf = output_view.get_buffer()
        output_scroll.set_child(output_view)
        content.append(output_scroll)

        # Command input
        cmd_entry = Gtk.Entry(placeholder_text="Type a CLI command...", hexpand=True)
        send_btn = Gtk.Button(icon_name="mail-send-symbolic")
        send_btn.add_css_class("suggested-action")

        input_box = Gtk.Box(spacing=8, margin_top=4)
        input_box.append(cmd_entry)
        input_box.append(send_btn)
        content.append(input_box)

        toolbar_view.set_content(content)
        dialog.set_child(toolbar_view)

        def append_output(text):
            end_iter = output_buf.get_end_iter()
            output_buf.insert(end_iter, text + "\n")
            GLib.idle_add(
                lambda: (
                    output_scroll.get_vadjustment().set_value(
                        output_scroll.get_vadjustment().get_upper()
                    )
                    or False
                )
            )

        def on_cli_response(cli_contact, message):
            """Handle CLI response messages from the room."""
            if cli_contact.public_key_hex == contact.public_key_hex and message.is_cli:
                append_output(message.text)

        # Register CLI response listener
        self._window.set_cli_response_callback(on_cli_response)

        def on_send(*args):
            cmd = cmd_entry.get_text().strip()
            if not cmd:
                return
            append_output(f"> {cmd}")
            cmd_entry.set_text("")
            from meshy.protocol import build_send_cli_command

            self._window.send_frame(build_send_cli_command(contact.public_key, cmd))

        send_btn.connect("clicked", on_send)
        cmd_entry.connect("activate", on_send)

        def on_dialog_closed(d):
            self._window.set_cli_response_callback(None)

        dialog.connect("closed", on_dialog_closed)

        # Send initial 'status' command
        append_output("> status")
        from meshy.protocol import build_send_cli_command

        self._window.send_frame(build_send_cli_command(contact.public_key, "status"))

        dialog.present(self._window)

    def _show_path_on_map(self, contact: Contact, hop_contacts: list, hop_prefixes: list = None):
        """Show the established path to a contact in a map dialog."""
        from meshy.views.map_view import (
            _MARKER_COLORS,
            _MARKER_ICONS,
            _SELF_COLOR,
            _UNKNOWN_COLOR,
            MapView,
            companion_too_far,
        )

        device = self._window.device_info
        self_lat = getattr(device, "latitude", None)
        self_lon = getattr(device, "longitude", None)
        has_self = self_lat and self_lon and (abs(self_lat) > 1e-6 or abs(self_lon) > 1e-6)

        # Build full hop list including unknowns: (lat, lon, contact, prefix, snr)
        raw_hops = []

        for i, hop_contact in enumerate(hop_contacts):
            prefix = hop_prefixes[i] if hop_prefixes and i < len(hop_prefixes) else None
            if hop_contact and hop_contact.has_location:
                raw_hops.append(
                    (hop_contact.latitude, hop_contact.longitude, hop_contact, prefix, None)
                )
            else:
                raw_hops.append((None, None, hop_contact, prefix, None))

        if contact.has_location:
            raw_hops.append((contact.latitude, contact.longitude, contact, None, None))

        if has_self and not companion_too_far(self_lat, self_lon, raw_hops):
            raw_hops.insert(0, (self_lat, self_lon, None, None, None))

        hops = MapView.interpolate_unknown_hops(raw_hops)

        if not any(h[0] is not None for h in hops):
            return

        dialog, map_widget, marker_layer, viewport = MapView.open_map_dialog(
            self._window, title=_("Path to {}").format(contact.name)
        )

        map_widget.get_map().remove_layer(marker_layer)
        path_layer = MapView.create_dashed_path_layer(viewport)
        map_widget.get_map().add_layer(path_layer)
        map_widget.get_map().add_layer(marker_layer)

        lats, lons = [], []
        hop_num = 0
        for lat, lon, hop_contact, prefix, _snr in hops:
            if lat is None:
                continue
            node = Shumate.Marker()
            node.set_location(lat, lon)
            path_layer.add_node(node)

            if hop_contact is None and prefix is None:
                # Self/device marker
                r, g, b = _SELF_COLOR
                icon_name = "network-wireless-symbolic"
                name = device.name or _("My Device")
                badge = ""
            elif hop_contact is None or not hop_contact.has_location:
                # Unknown/interpolated hop
                r, g, b = _UNKNOWN_COLOR
                icon_name = "network-server-symbolic"
                name = prefix or _("Unknown")
                hop_num += 1
                badge = str(hop_num)
            else:
                r, g, b = _MARKER_COLORS.get(hop_contact.type, (0.5, 0.5, 0.5))
                icon_name = _MARKER_ICONS.get(hop_contact.type, "avatar-default-symbolic")
                name = hop_contact.name
                hop_num += 1
                badge = str(hop_num)

            widget = MapView._build_marker_widget(r, g, b, icon_name, "", badge_text=badge)
            widget.set_tooltip_text(name)

            if hop_contact is not None:
                click = Gtk.GestureClick()
                click.connect(
                    "pressed", lambda g, n, x, y, c=hop_contact: self._show_contact_detail(c)
                )
                widget.add_controller(click)
            elif prefix is not None:
                click = Gtk.GestureClick()
                click.connect(
                    "pressed",
                    lambda g, n, x, y, w=widget, nm=name, la=lat, lo=lon: (
                        MapView.show_marker_popover(
                            w, nm, la, lo, contact_type=_("Unknown repeater"), interpolated=True
                        )
                    ),
                )
                widget.add_controller(click)

            hop_marker = Shumate.Marker()
            hop_marker.set_child(widget)
            hop_marker.set_location(lat, lon)
            marker_layer.add_marker(hop_marker)

            lats.append(lat)
            lons.append(lon)

        MapView.fit_viewport(viewport, lats, lons)
        dialog.present(self._window)

    def _show_on_map(self, contact: Contact):
        """Show a contact's location in a map dialog."""
        from meshy.views.map_view import _MARKER_COLORS, _MARKER_ICONS, MapView

        dialog, map_widget, marker_layer, viewport = MapView.open_map_dialog(
            self._window, title=contact.name
        )

        r, g, b = _MARKER_COLORS.get(contact.type, (0.5, 0.5, 0.5))
        icon_name = _MARKER_ICONS.get(contact.type, "avatar-default-symbolic")
        marker = MapView.create_marker(
            r, g, b, icon_name, contact.name, contact.latitude, contact.longitude
        )
        marker_layer.add_marker(marker)

        viewport.set_zoom_level(14)
        viewport.set_location(contact.latitude, contact.longitude)
        dialog.present(self._window)

    def _show_contact_detail(self, contact: Contact):
        """Show contact detail dialog."""
        from meshy.views.contact_detail_dialog import ContactDetailDialog

        dialog = ContactDetailDialog(self, contact)
        dialog.present(self._window)



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""MeshCore protocol - sync wrappers using meshcore library's packets and parsing logic.

Uses meshcore.packets for constants and follows meshcore.reader's parsing format
for field names and response handling.
"""

import hashlib
import io
import struct
from datetime import datetime

from meshy.meshcore_packets import CommandType, ControlType, PacketType
from meshy.smaz import try_decode_prefixed as _smaz_decode

# Re-export constants that the rest of the app uses
CMD_APP_START = CommandType.APP_START.value
CMD_DEVICE_QUERY = CommandType.DEVICE_QUERY.value
CMD_GET_CONTACTS = CommandType.GET_CONTACTS.value
CMD_GET_DEVICE_TIME = CommandType.GET_DEVICE_TIME.value
CMD_SET_DEVICE_TIME = CommandType.SET_DEVICE_TIME.value
CMD_SEND_SELF_ADVERT = CommandType.SEND_SELF_ADVERT.value
CMD_SET_ADVERT_NAME = CommandType.SET_ADVERT_NAME.value
CMD_ADD_UPDATE_CONTACT = CommandType.ADD_UPDATE_CONTACT.value
CMD_SYNC_NEXT_MESSAGE = CommandType.SYNC_NEXT_MESSAGE.value
CMD_SET_RADIO_PARAMS = CommandType.SET_RADIO_PARAMS.value
CMD_SET_RADIO_TX_POWER = CommandType.SET_RADIO_TX_POWER.value
CMD_RESET_PATH = CommandType.RESET_PATH.value
CMD_SET_ADVERT_LATLON = CommandType.SET_ADVERT_LATLON.value
CMD_REMOVE_CONTACT = CommandType.REMOVE_CONTACT.value
CMD_EXPORT_CONTACT = CommandType.EXPORT_CONTACT.value
CMD_IMPORT_CONTACT = CommandType.IMPORT_CONTACT.value
CMD_REBOOT = 19
CMD_GET_BATT_AND_STORAGE = CommandType.GET_BATT_AND_STORAGE.value
CMD_SEND_LOGIN = CommandType.SEND_LOGIN.value
CMD_SEND_STATUS_REQ = CommandType.SEND_STATUS_REQ.value
CMD_GET_CHANNEL = CommandType.GET_CHANNEL.value
CMD_SET_CHANNEL = CommandType.SET_CHANNEL.value
CMD_SEND_TRACE_PATH = CommandType.SEND_TRACE_PATH.value
CMD_SET_OTHER_PARAMS = CommandType.SET_OTHER_PARAMS.value
CMD_SEND_TXT_MSG = CommandType.SEND_TXT_MSG.value
CMD_SEND_CHANNEL_TXT_MSG = CommandType.SEND_CHANNEL_TXT_MSG.value
CMD_GET_AUTO_ADD_CONFIG = CommandType.GET_AUTOADD_CONFIG.value
CMD_SET_AUTO_ADD_CONFIG = CommandType.SET_AUTOADD_CONFIG.value
CMD_SEND_BINARY_REQ = CommandType.BINARY_REQ.value
CMD_GET_CUSTOM_VAR = CommandType.GET_CUSTOM_VARS.value
CMD_SET_CUSTOM_VAR = CommandType.SET_CUSTOM_VAR.value
CMD_GET_STATS = 56
CMD_SET_DEVICE_PIN = CommandType.SET_DEVICE_PIN.value
CMD_SET_PATH_HASH_MODE = CommandType.SET_PATH_HASH_MODE.value

# Response/push codes from meshcore.packets.PacketType
RESP_CODE_OK = PacketType.OK.value
RESP_CODE_ERR = PacketType.ERROR.value
RESP_CODE_CONTACTS_START = PacketType.CONTACT_START.value
RESP_CODE_CONTACT = PacketType.CONTACT.value
RESP_CODE_END_OF_CONTACTS = PacketType.CONTACT_END.value
RESP_CODE_SELF_INFO = PacketType.SELF_INFO.value
RESP_CODE_SENT = PacketType.MSG_SENT.value
RESP_CODE_CONTACT_MSG_RECV = PacketType.CONTACT_MSG_RECV.value
RESP_CODE_CHANNEL_MSG_RECV = PacketType.CHANNEL_MSG_RECV.value
RESP_CODE_CURR_TIME = PacketType.CURRENT_TIME.value
RESP_CODE_NO_MORE_MESSAGES = PacketType.NO_MORE_MSGS.value
RESP_CODE_EXPORT_CONTACT = PacketType.CONTACT_URI.value
RESP_CODE_BINARY_RESPONSE = PacketType.BINARY_RESPONSE.value
RESP_CODE_BATT_AND_STORAGE = PacketType.BATTERY.value
RESP_CODE_DEVICE_INFO = PacketType.DEVICE_INFO.value
RESP_CODE_CONTACT_MSG_RECV_V3 = 16
RESP_CODE_CHANNEL_MSG_RECV_V3 = 17
RESP_CODE_CHANNEL_INFO = PacketType.CHANNEL_INFO.value
RESP_CODE_AUTO_ADD_CONFIG = PacketType.AUTOADD_CONFIG.value
RESP_CODE_CHANNEL_DATA_RECV = PacketType.CHANNEL_DATA_RECV.value
RESP_CODE_DEFAULT_FLOOD_SCOPE = PacketType.DEFAULT_FLOOD_SCOPE.value
RESP_CODE_STATS = PacketType.STATS.value
RESP_CODE_ADVERT_PATH = PacketType.ADVERT_PATH.value

PUSH_CODE_ADVERT = PacketType.ADVERTISEMENT.value
PUSH_CODE_PATH_UPDATED = PacketType.PATH_UPDATE.value
PUSH_CODE_SEND_CONFIRMED = PacketType.ACK.value
PUSH_CODE_MSG_WAITING = PacketType.MESSAGES_WAITING.value
PUSH_CODE_NEW_ADVERT = PacketType.PUSH_CODE_NEW_ADVERT.value
PUSH_CODE_TRACE_DATA = PacketType.TRACE_DATA.value
PUSH_CODE_LOG_DATA = PacketType.LOG_DATA.value  # 0x88 - raw radio packet
PUSH_CODE_TELEMETRY_RESPONSE = PacketType.TELEMETRY_RESPONSE.value
PUSH_CODE_PATH_DISCOVERY_RESPONSE = PacketType.PATH_DISCOVERY_RESPONSE.value  # 0x8D
PUSH_CODE_LOGIN_SUCCESS = PacketType.LOGIN_SUCCESS.value  # 0x85
PUSH_CODE_LOGIN_FAILED = PacketType.LOGIN_FAILED.value  # 0x86
PUSH_CODE_STATUS_RESPONSE = PacketType.STATUS_RESPONSE.value  # 0x87

# Text types
TXT_TYPE_PLAIN = 0
TXT_TYPE_CLI_DATA = 1
TXT_TYPE_SIGNED = 2

# Contact types
ADV_TYPE_CHAT = 1
ADV_TYPE_REPEATER = 2
ADV_TYPE_ROOM = 3
ADV_TYPE_SENSOR = 4

# Sizes
PUB_KEY_SIZE = 32
MAX_PATH_SIZE = 64
MAX_NAME_SIZE = 32
MAX_FRAME_SIZE = 172
MAX_TEXT_PAYLOAD = 160  # firmware payload limit for text messages
APP_PROTOCOL_VERSION = 4

# Message overhead: cmd(1) + type(1) + attempt(1) + ts(4) + pubkey(6) + null(1) + safety(2)
_DM_OVERHEAD = 16
# Channel overhead: cmd(1) + type(1) + chan_idx(1) + ts(4) + null(1) + safety(2)
_CHAN_OVERHEAD = 10


def max_dm_text_bytes() -> int:
    """Max UTF-8 bytes for a direct message."""
    return min(MAX_FRAME_SIZE - _DM_OVERHEAD, MAX_TEXT_PAYLOAD)


def max_channel_text_bytes(sender_name: str) -> int:
    """Max UTF-8 bytes for a channel message (firmware prepends 'name: ')."""
    prefix_len = len(sender_name.encode("utf-8")) + 2  # ": " appended
    return min(MAX_TEXT_PAYLOAD - prefix_len, MAX_FRAME_SIZE - _CHAN_OVERHEAD)


# Binary request types
REQ_TYPE_STATUS = 0x01
REQ_TYPE_TELEMETRY = 0x03
REQ_TYPE_ACL = 0x05
REQ_TYPE_GET_NEIGHBORS = 0x06

# Control types (for zero-hop discovery)
CTRL_NODE_DISCOVER_REQ = ControlType.NODE_DISCOVER_REQ.value  # 0x80
CTRL_NODE_DISCOVER_RESP = ControlType.NODE_DISCOVER_RESP.value  # 0x90

# Response code for control data
RESP_CODE_CONTROL_DATA = PacketType.CONTROL_DATA.value  # 0x8E

CMD_SEND_CONTROL_DATA = CommandType.SEND_CONTROL_DATA.value  # 55

# Auto-add flags
AUTO_ADD_OVERWRITE_OLDEST = 1 << 0
AUTO_ADD_CHAT = 1 << 1
AUTO_ADD_REPEATER = 1 << 2
AUTO_ADD_ROOM_SERVER = 1 << 3
AUTO_ADD_SENSOR = 1 << 4

# Stats sub-types
STATS_TYPE_CORE = 0
STATS_TYPE_RADIO = 1
STATS_TYPE_PACKETS = 2

# Error codes from companion firmware (MyMesh.cpp)
ERROR_CODES = {
    0x01: "Unsupported command",
    0x02: "Not found",
    0x03: "Table full",
    0x04: "Bad state",
    0x05: "File I/O error",
    0x06: "Illegal argument",
}


def _now_ts() -> int:
    return int(datetime.now().timestamp())


def _decode_cstring(data: bytes) -> str:
    try:
        idx = data.index(0)
        return data[:idx].decode("utf-8", errors="replace")
    except ValueError:
        return data.decode("utf-8", errors="replace")


def _parse_path_length_byte(plen: int) -> tuple[int, int]:
    if plen == 255:
        return (-1, 1)
    return (plen & 0x3F, ((plen >> 6) & 0x03) + 1)


def _parse_snr(data: bytes) -> float:
    return int.from_bytes(data, byteorder="little", signed=True) / 4.0


def _parse_latlon(dbuf) -> tuple[float, float]:
    lat = int.from_bytes(dbuf.read(4), byteorder="little", signed=True) / 1e6
    lon = int.from_bytes(dbuf.read(4), byteorder="little", signed=True) / 1e6
    return (lat, lon)


def _build_scope_key(scope: str) -> bytes:
    if not scope or scope in ("0", "None", "*", ""):
        return bytes(16)
    if not scope.startswith("#"):
        scope = "#" + scope
    return hashlib.sha256(scope.encode("utf-8")).digest()[:16]


# ─── Frame Builders ───────────────────────────────────────────────
# Following meshcore_py commands/* patterns for frame construction


def build_app_start(app_name: str = "Meshy", app_version: int = APP_PROTOCOL_VERSION) -> bytes:
    return bytes([CMD_APP_START, app_version]) + bytes(6) + app_name.encode("utf-8") + b"\x00"


def build_device_query(app_version: int = APP_PROTOCOL_VERSION) -> bytes:
    return bytes([CMD_DEVICE_QUERY, app_version])


def build_get_contacts(since: int | None = None) -> bytes:
    frame = bytes([CMD_GET_CONTACTS])
    if since is not None:
        frame += struct.pack("<I", since)
    return frame


def build_get_device_time() -> bytes:
    return bytes([CMD_GET_DEVICE_TIME])


def build_set_device_time(timestamp: int | None = None) -> bytes:
    if timestamp is None:
        timestamp = _now_ts()
    return bytes([CMD_SET_DEVICE_TIME]) + struct.pack("<I", timestamp)


def build_get_batt_and_storage() -> bytes:
    return bytes([CMD_GET_BATT_AND_STORAGE])


def build_sync_next_message() -> bytes:
    return bytes([CMD_SYNC_NEXT_MESSAGE])


def build_send_text_msg(
    recipient_pub_key: bytes, text: str, attempt: int = 0, timestamp: int | None = None
) -> bytes:
    if timestamp is None:
        timestamp = _now_ts()
    frame = bytes([CMD_SEND_TXT_MSG, TXT_TYPE_PLAIN, min(attempt, 255)])
    frame += struct.pack("<I", timestamp)
    frame += recipient_pub_key[:6]
    frame += text.encode("utf-8") + b"\x00"
    return frame


def build_send_channel_text_msg(
    channel_index: int, text: str, timestamp: int | None = None
) -> bytes:
    if timestamp is None:
        timestamp = _now_ts()
    frame = bytes([CMD_SEND_CHANNEL_TXT_MSG, TXT_TYPE_PLAIN, channel_index])
    frame += struct.pack("<I", timestamp)
    frame += text.encode("utf-8") + b"\x00"
    return frame


def build_send_cli_command(
    repeater_pub_key: bytes, command: str, attempt: int = 0, timestamp: int | None = None
) -> bytes:
    if timestamp is None:
        timestamp = _now_ts()
    frame = bytes([CMD_SEND_TXT_MSG, TXT_TYPE_CLI_DATA, min(attempt, 255)])
    frame += struct.pack("<I", timestamp)
    frame += repeater_pub_key[:6]
    frame += command.encode("utf-8") + b"\x00"
    return frame


def build_send_self_advert(flood: bool = False) -> bytes:
    if flood:
        return bytes([CMD_SEND_SELF_ADVERT, 1])
    return bytes([CMD_SEND_SELF_ADVERT])


def build_set_advert_name(name: str) -> bytes:
    return bytes([CMD_SET_ADVERT_NAME]) + name.encode("utf-8")[: MAX_NAME_SIZE - 1]


def build_set_advert_latlon(lat: float, lon: float) -> bytes:
    frame = bytes([CMD_SET_ADVERT_LATLON])
    frame += struct.pack("<i", round(lat * 1e6))
    frame += struct.pack("<i", round(lon * 1e6))
    frame += struct.pack("<i", 0)  # altitude (reserved)
    return frame


def build_remove_contact(pub_key: bytes) -> bytes:
    return bytes([CMD_REMOVE_CONTACT]) + pub_key


def build_add_update_contact(
    pub_key: bytes,
    contact_type: int = ADV_TYPE_CHAT,
    flags: int = 0,
    path_len: int = 0,
    path: bytes = b"",
    name: str = "",
    lat: float | None = None,
    lon: float | None = None,
    last_modified: int | None = None,
) -> bytes:
    frame = bytes([CMD_ADD_UPDATE_CONTACT])
    frame += pub_key
    frame += bytes([contact_type, flags, path_len])
    path_padded = bytearray(MAX_PATH_SIZE)
    path_padded[: len(path)] = path[:MAX_PATH_SIZE]
    frame += bytes(path_padded)
    name_bytes = name.encode("utf-8")[: MAX_NAME_SIZE - 1]
    frame += name_bytes + b"\x00" * (MAX_NAME_SIZE - len(name_bytes))
    frame += struct.pack("<I", _now_ts())
    has_location = lat is not None and lon is not None
    if has_location or last_modified is not None:
        frame += struct.pack("<i", round(lat * 1e6) if has_location else 0)
        frame += struct.pack("<i", round(lon * 1e6) if has_location else 0)
        if last_modified is not None:
            frame += struct.pack("<I", last_modified)
    return frame


def build_reset_path(pub_key: bytes) -> bytes:
    return bytes([CMD_RESET_PATH]) + pub_key


def build_export_contact(pub_key: bytes) -> bytes:
    return bytes([CMD_EXPORT_CONTACT]) + pub_key


def build_import_contact(contact_frame: bytes) -> bytes:
    return bytes([CMD_IMPORT_CONTACT]) + contact_frame


def build_get_channel(channel_index: int) -> bytes:
    return bytes([CMD_GET_CHANNEL, channel_index])


def build_set_channel(channel_index: int, name: str, psk: bytes) -> bytes:
    name_bytes = name.encode("utf-8")[:32]
    name_bytes = name_bytes.ljust(32, b"\x00")
    psk_padded = bytearray(16)
    psk_padded[: len(psk)] = psk[:16]
    return bytes([CMD_SET_CHANNEL, channel_index]) + name_bytes + bytes(psk_padded)


def build_set_radio_params(
    freq_hz: int, bw_hz: int, sf: int, cr: int, client_repeat: bool | None = None
) -> bytes:
    # freq_hz and bw_hz are in kHz (matching meshcore_py: freq*1000, bw*1000)
    frame = bytes([CMD_SET_RADIO_PARAMS])
    frame += struct.pack("<I", freq_hz)
    frame += struct.pack("<I", bw_hz)
    frame += bytes([sf, cr])
    if client_repeat is not None:
        frame += bytes([1 if client_repeat else 0])
    return frame


def build_set_radio_tx_power(power_dbm: int) -> bytes:
    return bytes([CMD_SET_RADIO_TX_POWER]) + struct.pack("<i", power_dbm)[:4]


def build_reboot() -> bytes:
    return bytes([CMD_REBOOT]) + b"reboot"


def build_factory_reset() -> bytes:
    return bytes([CommandType.FACTORY_RESET.value]) + b"reset"


def build_set_path_hash_mode(mode: int) -> bytes:
    """Set path hash mode: 0=1-byte, 1=2-byte, 2=3-byte."""
    return bytes([CMD_SET_PATH_HASH_MODE, 0, mode])


def build_set_device_pin(pin: int) -> bytes:
    return bytes([CMD_SET_DEVICE_PIN]) + struct.pack("<I", pin)


def build_set_other_params(
    manual_add_contacts: int = 0,
    telemetry_flags: int = 0,
    advert_loc_policy: int = 0,
    multi_acks: int = 0,
) -> bytes:
    """Set other params.

    Format: [0x26][manual_add_contacts][telemetry_mode][advert_loc_policy][multi_acks]
    """
    return bytes(
        [CMD_SET_OTHER_PARAMS, manual_add_contacts, telemetry_flags, advert_loc_policy, multi_acks]
    )


def build_get_custom_vars() -> bytes:
    return bytes([CMD_GET_CUSTOM_VAR])


def build_set_custom_var(key_value: str) -> bytes:
    """Set a custom variable. Format: 'key:value' (e.g. 'gps:1')."""
    return bytes([CMD_SET_CUSTOM_VAR]) + key_value.encode("utf-8") + b"\x00"


RESP_CODE_CUSTOM_VARS = PacketType.CUSTOM_VARS.value


def parse_custom_vars(data: bytes) -> dict:
    """Parse CUSTOM_VARS response. Format: 'key1:val1,key2:val2,...'"""
    if len(data) < 2 or data[0] != RESP_CODE_CUSTOM_VARS:
        return {}
    raw = data[1:].decode("utf-8", "ignore").rstrip("\x00")
    if not raw:
        return {}
    result = {}
    for pair in raw.split(","):
        parts = pair.split(":", 1)
        if len(parts) == 2:
            result[parts[0]] = parts[1]
    return result


def build_set_flood_scope(scope: str = "", unscoped: bool = False) -> bytes:
    """Set session flood scope (cmd 54). Empty string or '*' to disable, '#name' for hashtag scope.

    When *unscoped* is True (fw v12+), sends with byte[1]=1 to explicitly
    override the default scope for this message.
    """
    cmd = CommandType.SET_FLOOD_SCOPE.value
    flags = 1 if unscoped else 0
    return bytes([cmd, flags]) + _build_scope_key(scope)


def build_set_default_flood_scope(scope: str = "") -> bytes:
    """Set persistent default flood scope (cmd 63). Empty to clear, '#name' for hashtag scope."""
    cmd = CommandType.SET_DEFAULT_FLOOD_SCOPE.value
    scope_key = _build_scope_key(scope)
    if scope_key == bytes(16):
        return bytes([cmd])
    if not scope.startswith("#"):
        scope = "#" + scope
    scope_name = scope.encode("utf-8")[:31].ljust(31, b"\x00")
    return bytes([cmd]) + scope_name + scope_key


def build_get_advert_path(pub_key: bytes) -> bytes:
    """Get the advertisement path cached for a contact. Uses full 32-byte key."""
    return bytes([CommandType.GET_ADVERT_PATH.value, 0]) + pub_key[:PUB_KEY_SIZE]


def build_path_discovery(pub_key: bytes) -> bytes:
    """Send a path discovery request for a contact."""
    return bytes([CommandType.PATH_DISCOVERY.value, 0]) + pub_key


def parse_advert_path_response(data: bytes) -> dict | None:
    """Parse RESP_CODE_ADVERT_PATH (0x16).
    Format: [code][timestamp×4 LE][path_len_byte][path_bytes...]
    """
    if len(data) < 6 or data[0] != PacketType.ADVERT_PATH.value:
        return None
    timestamp = struct.unpack_from("<I", data, 1)[0]
    path_len_byte = data[5]
    if path_len_byte == 255:
        return {"path_length": -1, "path": b"", "path_hash_size": 1, "timestamp": timestamp}
    hop_count = path_len_byte & 0x3F
    hash_size = ((path_len_byte >> 6) & 0x03) + 1
    path_byte_len = hop_count * hash_size
    path_data = data[6 : 6 + path_byte_len] if path_byte_len > 0 else b""
    return {
        "path_length": hop_count,
        "path_hash_size": hash_size,
        "path": path_data,
        "timestamp": timestamp,
        "hops": [
            path_data[i : i + hash_size].hex().upper() for i in range(0, len(path_data), hash_size)
        ]
        if path_data
        else [],
    }


def build_trace_path(tag: int, path_bytes: bytes, flags: int = 0) -> bytes:
    """Build a SEND_TRACE_PATH request.
    Format: [cmd][tag×4 LE][auth×4 LE][flag][path_hops...]
    flags encodes path hash size: 0=1-byte, 1=2-byte, 2=4-byte, 3=8-byte.
    """
    frame = bytes([CMD_SEND_TRACE_PATH])
    frame += struct.pack("<I", tag)
    frame += struct.pack("<I", 0)  # auth
    frame += bytes([flags])
    frame += path_bytes
    return frame


def parse_trace_response(data: bytes) -> dict | None:
    """Parse PUSH_CODE_TRACE_DATA (0x89) response.
    Format: [code][reserved][path_byte_len][flags][tag×4][auth×4][path_data...][snr_values...]
    flags lower 2 bits encode hash size: 0→1B, 1→2B, 2→4B, 3→8B.
    """
    if len(data) < 12 or data[0] != PUSH_CODE_TRACE_DATA:
        return None
    path_byte_len = data[2]
    flags = data[3]
    path_hash_size = 1 << (flags & 3)
    hop_count = path_byte_len // path_hash_size if path_hash_size else path_byte_len
    tag = data[4:8]
    # path data starts at offset 12
    offset = 12
    if offset + path_byte_len > len(data):
        return None
    path_data = data[offset : offset + path_byte_len]
    snr_raw = data[offset + path_byte_len :]
    snr_values = [_parse_snr(bytes([b])) for b in snr_raw]
    # Split path into per-hop hashes
    path_hops = [
        path_data[i : i + path_hash_size]
        for i in range(0, hop_count * path_hash_size, path_hash_size)
    ]
    return {
        "tag": tag,
        "path_hops": path_hops,
        "path_hash_size": path_hash_size,
        "snr_values": snr_values,
    }


def build_send_login(recipient_pub_key: bytes, password: str) -> bytes:
    return bytes([CMD_SEND_LOGIN]) + recipient_pub_key + password.encode("utf-8") + b"\x00"


def build_send_status_req(recipient_pub_key: bytes) -> bytes:
    return bytes([CMD_SEND_STATUS_REQ]) + recipient_pub_key


def build_get_auto_add_config() -> bytes:
    return bytes([CMD_GET_AUTO_ADD_CONFIG])


def build_set_auto_add_config(
    chat: bool = False,
    repeater: bool = False,
    room: bool = False,
    sensor: bool = False,
    overwrite_oldest: bool = False,
    max_hops: int = 0,
) -> bytes:
    flags = 0
    if chat:
        flags |= AUTO_ADD_CHAT
    if repeater:
        flags |= AUTO_ADD_REPEATER
    if room:
        flags |= AUTO_ADD_ROOM_SERVER
    if sensor:
        flags |= AUTO_ADD_SENSOR
    if overwrite_oldest:
        flags |= AUTO_ADD_OVERWRITE_OLDEST
    return bytes([CMD_SET_AUTO_ADD_CONFIG, flags, min(max_hops, 64)])


def build_send_discover_req(
    type_filter: int = 0xFF, prefix_only: bool = True, tag: int | None = None
) -> bytes:
    """Send a zero-hop NODE_DISCOVER_REQ control packet.

    Args:
        type_filter: bitmask of ADV_TYPE_* bits (0xFF = all types)
        prefix_only: if True, responders send 8-byte key prefix; if False, full 32-byte key
        tag: random correlation tag (auto-generated if None)
    """
    import random

    if tag is None:
        tag = random.randint(0, 0xFFFFFFFF)
    flags = CTRL_NODE_DISCOVER_REQ  # 0x80 = upper nibble is sub_type 8
    if prefix_only:
        flags |= 0x01  # bit 0 = prefix_only
    frame = bytes([CMD_SEND_CONTROL_DATA, flags, type_filter])
    frame += struct.pack("<I", tag)
    return frame


def parse_discover_response(data: bytes) -> dict | None:
    """Parse a NODE_DISCOVER_RESP from a CONTROL_DATA push (0x8E).

    Companion protocol format:
      0x8E + 2 prefix bytes + path_len(1) + path + flags(1) + snr(1) + tag(4) + pubkey
    """
    if len(data) < 8:
        return None
    # data[0] = 0x8E, data[1:3] = prefix bytes, data[3] = path_length
    offset = 3  # skip code + 2 prefix bytes
    path_len_byte = data[offset]
    hop_count = path_len_byte & 0x3F
    hash_size = ((path_len_byte >> 6) & 0x03) + 1
    offset += 1
    offset += hop_count * hash_size  # skip path hops

    if offset + 6 > len(data):  # need flags + snr + tag(4)
        return None

    flags = data[offset]
    sub_type = (flags >> 4) & 0x0F
    if sub_type != 0x09:  # DISCOVER_RESP sub_type = 9
        return None

    node_type = flags & 0x0F
    snr = _parse_snr(data[offset + 1 : offset + 2])
    tag = int.from_bytes(data[offset + 2 : offset + 6], "little", signed=False)
    pub_key_hex = data[offset + 6 :].hex()

    return {
        "node_type": node_type,
        "snr": snr,
        "tag": tag,
        "pub_key_hex": pub_key_hex,
    }


def build_get_neighbors(
    repeater_pub_key: bytes, max_results: int = 10, offset: int = 0, key_len: int = 4
) -> bytes:
    """Build a binary request to get neighbors from a repeater."""
    frame = bytes([CMD_SEND_BINARY_REQ])
    frame += repeater_pub_key
    frame += bytes(
        [
            REQ_TYPE_GET_NEIGHBORS,
            0x00,  # version/reserved
            max_results,  # max number of neighbors
        ]
    )
    frame += struct.pack("<H", offset)  # offset (16-bit LE)
    frame += bytes(
        [
            0x00,  # order by
            key_len,  # public key prefix length
        ]
    )
    return frame


def build_get_acl(repeater_pub_key: bytes) -> bytes:
    """Build a binary request to get ACL from a repeater."""
    frame = bytes([CMD_SEND_BINARY_REQ])
    frame += repeater_pub_key
    frame += bytes([REQ_TYPE_ACL, 0x00, 0x00])
    return frame


def build_get_telemetry(repeater_pub_key: bytes) -> bytes:
    """Build a binary request to get telemetry from a repeater/sensor."""
    frame = bytes([CMD_SEND_BINARY_REQ])
    frame += repeater_pub_key
    frame += bytes([REQ_TYPE_TELEMETRY])
    return frame


def build_get_telemetry_chat(pub_key: bytes = None) -> bytes:
    """Build a telemetry request for a chat companion node.

    Uses CMD_SEND_TELEMETRY_REQ (0x27) instead of CMD_SEND_BINARY_REQ.
    Response arrives as PUSH_CODE_TELEMETRY_RESPONSE (0x8B).
    If pub_key is None, requests self-telemetry from the connected device.
    """
    cmd = CommandType.SEND_TELEMETRY_REQ.value
    frame = bytes([cmd])
    if pub_key and len(pub_key) >= 32:
        frame += bytes(3)  # reserved
        frame += pub_key[:32]
    else:
        frame += bytes(3)  # reserved (self-telemetry)
    return frame


ANON_REQ_REGIONS = 0x01


def build_anon_req_regions(pub_key: bytes, path_length: int, path: bytes) -> bytes:
    """Build an anonymous request for regions from a repeater.

    Format: [CMD_SEND_ANON_REQ][pub_key_32][req_type][path_len][path_reversed]
    """
    cmd = CommandType.SEND_ANON_REQ.value
    frame = bytes([cmd])
    frame += pub_key[:32]
    frame += bytes([ANON_REQ_REGIONS])
    frame += bytes([path_length])
    frame += bytes(reversed(path[:path_length])) if path_length > 0 else b""
    return frame


def parse_status_response(data: bytes) -> dict | None:
    """Parse PUSH_CODE_STATUS_RESPONSE (0x87) — 52/56-byte binary status.
    Frame: [code(1)][reserved(1)][sender_prefix(6)][status_data(52+)]
    """
    if len(data) < 60:  # 8 header + 52 payload
        return None
    # Skip 8 bytes: code(1) + reserved(1) + sender_prefix(6)
    d = data[8:]
    return {
        "battery_mv": struct.unpack("<H", d[0:2])[0],
        "queue_len": struct.unpack("<H", d[2:4])[0],
        "noise_floor": struct.unpack("<h", d[4:6])[0],
        "last_rssi": struct.unpack("<h", d[6:8])[0],
        "packets_recv": struct.unpack("<I", d[8:12])[0],
        "packets_sent": struct.unpack("<I", d[12:16])[0],
        "tx_air_secs": struct.unpack("<I", d[16:20])[0],
        "uptime_secs": struct.unpack("<I", d[20:24])[0],
        "flood_tx": struct.unpack("<I", d[24:28])[0],
        "direct_tx": struct.unpack("<I", d[28:32])[0],
        "flood_rx": struct.unpack("<I", d[32:36])[0],
        "direct_rx": struct.unpack("<I", d[36:40])[0],
        "error_events": struct.unpack("<H", d[40:42])[0],
        "last_snr": struct.unpack("<h", d[42:44])[0] / 4.0,
        "dup_direct": struct.unpack("<H", d[44:46])[0] if len(d) >= 46 else 0,
        "dup_flood": struct.unpack("<H", d[46:48])[0] if len(d) >= 48 else 0,
        "rx_air_secs": struct.unpack("<I", d[48:52])[0] if len(d) >= 52 else 0,
        "recv_errors": struct.unpack("<I", d[52:56])[0] if len(d) >= 56 else None,
    }


def parse_neighbors_response(data: bytes, key_len: int = 4) -> dict:
    """Parse a neighbors binary response (0x8C).
    Frame: [code(1)][reserved(1)][tag(4)][total_count(2 LE)][results_count(2 LE)][entries...]
    Each entry: key_prefix(key_len) + last_heard(4 LE) + snr(int8).
    Returns dict with 'total', 'neighbors' list.
    """
    if len(data) < 10:
        return {"total": 0, "neighbors": []}
    # Skip 6-byte header: code(1) + reserved(1) + tag(4)
    total_count = struct.unpack("<H", data[6:8])[0]
    results_count = struct.unpack("<H", data[8:10])[0]
    buf = data[10:]
    neighbors = []
    entry_size = key_len + 4 + 1
    for _ in range(results_count):
        if len(buf) < entry_size:
            break
        key_prefix = buf[:key_len].hex()
        last_heard = struct.unpack("<I", buf[key_len : key_len + 4])[0]
        snr = _parse_snr(buf[key_len + 4 : key_len + 5])
        neighbors.append(
            {
                "pub_key_hex": key_prefix,
                "last_heard": last_heard,
                "snr": snr,
            }
        )
        buf = buf[entry_size:]
    return {"total": total_count, "neighbors": neighbors}


def build_get_stats(sub_type: int = STATS_TYPE_CORE) -> bytes:
    """Build a stats request. sub_type: 0=Core, 1=Radio, 2=Packets."""
    return bytes([CMD_GET_STATS, sub_type])


def parse_stats_response(data: bytes) -> dict | None:
    """Parse stats response (code 0x18). Byte 1 is sub-type."""
    if len(data) < 4:
        return None
    sub_type = data[1]
    result = {"sub_type": sub_type}

    if sub_type == STATS_TYPE_CORE and len(data) >= 11:
        result["battery_mv"] = int.from_bytes(data[2:4], "little", signed=False)
        result["uptime_secs"] = int.from_bytes(data[4:8], "little", signed=False)
        result["errors"] = int.from_bytes(data[8:10], "little", signed=False)
        result["queue_len"] = data[10]

    elif sub_type == STATS_TYPE_RADIO and len(data) >= 14:
        result["noise_floor"] = int.from_bytes(data[2:4], "little", signed=True)
        result["last_rssi"] = int.from_bytes(data[4:5], "little", signed=True)
        result["last_snr"] = _parse_snr(data[5:6])
        result["tx_air_secs"] = int.from_bytes(data[6:10], "little", signed=False)
        result["rx_air_secs"] = int.from_bytes(data[10:14], "little", signed=False)

    elif sub_type == STATS_TYPE_PACKETS and len(data) >= 26:
        result["recv"] = int.from_bytes(data[2:6], "little", signed=False)
        result["sent"] = int.from_bytes(data[6:10], "little", signed=False)
        result["flood_tx"] = int.from_bytes(data[10:14], "little", signed=False)
        result["direct_tx"] = int.from_bytes(data[14:18], "little", signed=False)
        result["flood_rx"] = int.from_bytes(data[18:22], "little", signed=False)
        result["direct_rx"] = int.from_bytes(data[22:26], "little", signed=False)
        if len(data) >= 30:
            result["recv_errors"] = int.from_bytes(data[26:30], "little", signed=False)

    return result


def parse_error_response(data: bytes) -> dict | None:
    """Parse error response (code 0x01). Byte 1 is error code."""
    if len(data) < 2:
        return {"code": 0, "message": "Unknown error"}
    code = data[1]
    return {
        "code": code,
        "message": ERROR_CODES.get(code, f"Unknown error (0x{code:02x})"),
    }


# ─── Frame Parsers ────────────────────────────────────────────────
# Sync parsing following meshcore.reader.MessageReader.handle_rx field names


def identify_response(data: bytes) -> int | None:
    if not data:
        return None
    return data[0]


def parse_contact_frame(data: bytes) -> dict | None:
    """Parse contact frame, matching meshcore reader field names."""
    if len(data) < 100:
        return None
    code = data[0]
    if code not in (RESP_CODE_CONTACT, PUSH_CODE_NEW_ADVERT):
        return None
    dbuf = io.BytesIO(data[1:])
    c = {}
    c["public_key"] = dbuf.read(32).hex()
    c["type"] = dbuf.read(1)[0]
    c["flags"] = dbuf.read(1)[0]
    plen = int.from_bytes(dbuf.read(1), signed=False, byteorder="little")
    c["out_path_len"], c["path_hash_size"] = _parse_path_length_byte(plen)
    raw_path = dbuf.read(64)
    # Extract meaningful path bytes based on hop count and hash size
    hop_count = max(c["out_path_len"], 0)
    hash_size = c["path_hash_size"]
    path_byte_len = hop_count * hash_size
    c["out_path"] = raw_path[:path_byte_len].hex() if path_byte_len > 0 else ""
    c["adv_name"] = dbuf.read(32).decode("utf-8", "ignore").replace("\0", "")
    c["last_advert"] = int.from_bytes(dbuf.read(4), byteorder="little")
    c["adv_lat"], c["adv_lon"] = _parse_latlon(dbuf)
    lastmod_bytes = dbuf.read(4)
    if lastmod_bytes and len(lastmod_bytes) == 4:
        c["lastmod"] = int.from_bytes(lastmod_bytes, byteorder="little")
    return c


def parse_self_info(data: bytes) -> dict | None:
    """Parse SELF_INFO, matching meshcore reader field names."""
    if len(data) < 36 or data[0] != RESP_CODE_SELF_INFO:
        return None
    dbuf = io.BytesIO(data[1:])
    si = {}
    si["adv_type"] = dbuf.read(1)[0]
    si["tx_power"] = int.from_bytes(dbuf.read(1), signed=True, byteorder="little")
    si["max_tx_power"] = dbuf.read(1)[0]
    si["public_key"] = dbuf.read(32).hex()
    si["adv_lat"], si["adv_lon"] = _parse_latlon(dbuf)
    si["multi_acks"] = dbuf.read(1)[0]
    si["adv_loc_policy"] = dbuf.read(1)[0]
    telemetry_mode = dbuf.read(1)[0]
    si["telemetry_mode_env"] = (telemetry_mode >> 4) & 0b11
    si["telemetry_mode_loc"] = (telemetry_mode >> 2) & 0b11
    si["telemetry_mode_base"] = telemetry_mode & 0b11
    si["manual_add_contacts"] = dbuf.read(1)[0] > 0
    si["radio_freq"] = int.from_bytes(dbuf.read(4), byteorder="little") / 1000
    si["radio_bw"] = int.from_bytes(dbuf.read(4), byteorder="little") / 1000
    si["radio_sf"] = dbuf.read(1)[0]
    si["radio_cr"] = dbuf.read(1)[0]
    si["name"] = dbuf.read().decode("utf-8", "ignore").rstrip("\x00")
    return si


def parse_device_info(data: bytes) -> dict | None:
    """Parse DEVICE_INFO, matching meshcore reader field names."""
    if len(data) < 4 or data[0] != RESP_CODE_DEVICE_INFO:
        return None
    dbuf = io.BytesIO(data[1:])
    res = {}
    fw_ver = dbuf.read(1)[0]
    res["fw_ver"] = fw_ver
    if fw_ver >= 3:
        res["max_contacts"] = dbuf.read(1)[0] * 2
        res["max_channels"] = dbuf.read(1)[0]
        res["ble_pin"] = int.from_bytes(dbuf.read(4), byteorder="little")
        res["fw_build"] = dbuf.read(12).decode("utf-8", "ignore").replace("\0", "")
        res["model"] = dbuf.read(40).decode("utf-8", "ignore").replace("\0", "")
        res["ver"] = dbuf.read(20).decode("utf-8", "ignore").replace("\0", "")
    if fw_ver >= 9:
        rpt = dbuf.read(1)
        if len(rpt) > 0:
            res["repeat"] = rpt[0] != 0
    if fw_ver >= 10:
        phm = dbuf.read(1)
        if len(phm) > 0:
            res["path_hash_mode"] = phm[0]
    return res


def parse_batt_and_storage(data: bytes) -> dict | None:
    """Parse battery/storage, matching meshcore reader field names."""
    if len(data) < 3 or data[0] != RESP_CODE_BATT_AND_STORAGE:
        return None
    dbuf = io.BytesIO(data[1:])
    result = {}
    result["level"] = int.from_bytes(dbuf.read(2), byteorder="little")
    if len(data) > 3:
        result["used_kb"] = int.from_bytes(dbuf.read(4), byteorder="little")
        result["total_kb"] = int.from_bytes(dbuf.read(4), byteorder="little")
    return result


def parse_contact_message(data: bytes) -> dict | None:
    """Parse incoming DM, matching meshcore reader field names."""
    if len(data) < 10:
        return None
    code = data[0]
    if code not in (RESP_CODE_CONTACT_MSG_RECV, RESP_CODE_CONTACT_MSG_RECV_V3):
        return None
    dbuf = io.BytesIO(data[1:])
    res = {}
    res["type"] = "PRIV"
    if code == RESP_CODE_CONTACT_MSG_RECV_V3:
        res["SNR"] = _parse_snr(dbuf.read(1))
        dbuf.read(2)  # reserved
    res["pubkey_prefix"] = dbuf.read(6).hex()
    plen = dbuf.read(1)[0]
    res["path_len"], res["path_hash_size"] = _parse_path_length_byte(plen)
    txt_type = dbuf.read(1)[0]
    res["txt_type"] = txt_type
    res["sender_timestamp"] = int.from_bytes(dbuf.read(4), byteorder="little")
    if txt_type == 2:
        res["signature"] = dbuf.read(4).hex()
    res["text"] = _smaz_decode(dbuf.read().decode("utf-8", "ignore").rstrip("\x00"))
    return res


def parse_channel_message(data: bytes) -> dict | None:
    """Parse incoming channel message, matching meshcore reader field names."""
    if len(data) < 8:
        return None
    code = data[0]
    if code not in (RESP_CODE_CHANNEL_MSG_RECV, RESP_CODE_CHANNEL_MSG_RECV_V3):
        return None
    dbuf = io.BytesIO(data[1:])
    res = {}
    res["type"] = "CHAN"
    if code == RESP_CODE_CHANNEL_MSG_RECV_V3:
        res["SNR"] = _parse_snr(dbuf.read(1))
        dbuf.read(2)  # reserved
    res["channel_idx"] = dbuf.read(1)[0]
    plen = dbuf.read(1)[0]
    res["path_len"], res["path_hash_size"] = _parse_path_length_byte(plen)
    res["txt_type"] = dbuf.read(1)[0]
    res["path_bytes"] = ""  # Path bytes come via LOG_DATA (0x88), not here
    res["sender_timestamp"] = int.from_bytes(dbuf.read(4), byteorder="little", signed=False)
    text = dbuf.read().strip(b"\0")
    res["text"] = text.decode("utf-8", "ignore")
    # Channel messages have format "name: message"
    if ": " in res["text"]:
        res["sender_name"], res["msg_text"] = res["text"].split(": ", 1)
    else:
        res["sender_name"] = ""
        res["msg_text"] = res["text"]
    res["msg_text"] = _smaz_decode(res["msg_text"])
    return res


def parse_channel_data(data: bytes) -> dict | None:
    """Parse CHANNEL_DATA_RECV (binary data on group channels, companion v1.15.0+).

    Frame: [code][SNR(i8)][reserved(2)][chan_idx][path_len][data_type(u16LE)][data_len][payload]
    """
    if len(data) < 9 or data[0] != RESP_CODE_CHANNEL_DATA_RECV:
        return None
    dbuf = io.BytesIO(data[1:])
    res = {}
    res["SNR"] = _parse_snr(dbuf.read(1))
    dbuf.read(2)  # reserved
    res["channel_idx"] = dbuf.read(1)[0]
    plen = dbuf.read(1)[0]
    res["path_len"], res["path_hash_size"] = _parse_path_length_byte(plen)
    res["data_type"] = int.from_bytes(dbuf.read(2), byteorder="little")
    res["data_len"] = dbuf.read(1)[0]
    res["payload"] = dbuf.read(res["data_len"]).hex()
    return res


def parse_channel_info(data: bytes) -> dict | None:
    """Parse channel info, matching meshcore reader field names."""
    if len(data) < 50 or data[0] != RESP_CODE_CHANNEL_INFO:
        return None
    dbuf = io.BytesIO(data[1:])
    res = {}
    res["channel_idx"] = dbuf.read(1)[0]
    name_bytes = dbuf.read(32)
    null_pos = name_bytes.find(0)
    if null_pos >= 0:
        res["channel_name"] = name_bytes[:null_pos].decode("utf-8", "ignore")
    else:
        res["channel_name"] = name_bytes.decode("utf-8", "ignore")
    res["channel_secret"] = dbuf.read(16)
    return res


def parse_sent_response(data: bytes) -> dict | None:
    """Parse RESP_CODE_SENT, matching meshcore reader field names."""
    if len(data) < 2 or data[0] != RESP_CODE_SENT:
        return None
    dbuf = io.BytesIO(data[1:])
    res = {}
    res["type"] = dbuf.read(1)[0]
    res["expected_ack"] = dbuf.read(4)
    res["suggested_timeout"] = int.from_bytes(dbuf.read(4), byteorder="little")
    return res


def parse_push_send_confirmed(data: bytes) -> dict | None:
    """Parse ACK push."""
    if len(data) < 5 or data[0] != PUSH_CODE_SEND_CONFIRMED:
        return None
    res = {}
    res["code"] = data[1:5].hex()
    if len(data) >= 9:
        res["trip_time_ms"] = struct.unpack_from("<I", data, 5)[0]
    return res


def parse_auto_add_config(data: bytes) -> dict | None:
    """Parse auto-add config response."""
    if len(data) < 2 or data[0] != RESP_CODE_AUTO_ADD_CONFIG:
        return None
    config = data[1]
    max_hops = data[2] if len(data) >= 3 else 0
    return {"config": config, "max_hops": max_hops}


def parse_default_flood_scope(data: bytes) -> dict | None:
    """Parse DEFAULT_FLOOD_SCOPE response (PacketType 28).

    48-byte frame = scope set: [code][scope_name(31)][scope_key(16)].
    1-byte frame = no scope configured.
    """
    if not data or data[0] != RESP_CODE_DEFAULT_FLOOD_SCOPE:
        return None
    res = {}
    if len(data) >= 48:
        res["scope_name"] = data[1:32].decode("utf-8", "ignore").replace("\0", "")
        res["scope_key"] = data[32:48].hex()
    return res


def parse_curr_time(data: bytes) -> int | None:
    if len(data) < 5 or data[0] != RESP_CODE_CURR_TIME:
        return None
    return struct.unpack_from("<I", data, 1)[0]


# ─── Cayenne LPP telemetry parsing ──────────────────────────

LPP_DIGITAL_INPUT = 0  # 1 byte
LPP_DIGITAL_OUTPUT = 1  # 1 byte
LPP_ANALOG_INPUT = 2  # 2 bytes, 0.01 signed
LPP_ANALOG_OUTPUT = 3  # 2 bytes, 0.01 signed
LPP_GENERIC_SENSOR = 100  # 4 bytes, unsigned
LPP_LUMINOSITY = 101  # 2 bytes, 1 lux unsigned
LPP_PRESENCE = 102  # 1 byte, bool
LPP_TEMPERATURE = 103  # 2 bytes, 0.1°C signed
LPP_HUMIDITY = 104  # 1 byte, 0.5% unsigned
LPP_ACCELEROMETER = 113  # 6 bytes, 2 per axis, 0.001G
LPP_PRESSURE = 115  # 2 bytes, 0.1 hPa unsigned
LPP_VOLTAGE = 116  # 2 bytes, 0.01V unsigned
LPP_CURRENT = 117  # 2 bytes, 0.001A unsigned
LPP_FREQUENCY = 118  # 4 bytes, 1Hz unsigned
LPP_PERCENTAGE = 120  # 1 byte, 1-100%
LPP_ALTITUDE = 121  # 2 bytes, 1m signed
LPP_CONCENTRATION = 125  # 2 bytes, 1 ppm unsigned
LPP_POWER = 128  # 2 bytes, 1W unsigned
LPP_DISTANCE = 130  # 4 bytes, 0.001m unsigned
LPP_ENERGY = 131  # 4 bytes, 0.001kWh unsigned
LPP_DIRECTION = 132  # 2 bytes, 1deg unsigned
LPP_UNIX_TIME = 133  # 4 bytes, unsigned
LPP_GYROMETER = 134  # 6 bytes, 2 per axis, 0.01°/s
LPP_COLOUR = 135  # 3 bytes, RGB
LPP_GPS = 136  # 3x3 bytes (lat, lon, alt)
LPP_SWITCH = 142  # 1 byte, 0/1

LPP_TYPE_NAMES = {
    LPP_DIGITAL_INPUT: "Digital Input",
    LPP_DIGITAL_OUTPUT: "Digital Output",
    LPP_ANALOG_INPUT: "Analog Input",
    LPP_ANALOG_OUTPUT: "Analog Output",
    LPP_GENERIC_SENSOR: "Generic Sensor",
    LPP_LUMINOSITY: "Luminosity",
    LPP_PRESENCE: "Presence",
    LPP_TEMPERATURE: "Temperature",
    LPP_HUMIDITY: "Humidity",
    LPP_ACCELEROMETER: "Accelerometer",
    LPP_PRESSURE: "Pressure",
    LPP_VOLTAGE: "Voltage",
    LPP_CURRENT: "Current",
    LPP_FREQUENCY: "Frequency",
    LPP_PERCENTAGE: "Percentage",
    LPP_ALTITUDE: "Altitude",
    LPP_CONCENTRATION: "Concentration",
    LPP_POWER: "Power",
    LPP_DISTANCE: "Distance",
    LPP_ENERGY: "Energy",
    LPP_DIRECTION: "Direction",
    LPP_UNIX_TIME: "Time",
    LPP_GYROMETER: "Gyrometer",
    LPP_COLOUR: "Colour",
    LPP_GPS: "GPS",
    LPP_SWITCH: "Switch",
}

LPP_TYPE_UNITS = {
    LPP_ANALOG_INPUT: "V",
    LPP_ANALOG_OUTPUT: "V",
    LPP_LUMINOSITY: "lux",
    LPP_TEMPERATURE: "°C",
    LPP_HUMIDITY: "%",
    LPP_ACCELEROMETER: "G",
    LPP_PRESSURE: "hPa",
    LPP_VOLTAGE: "V",
    LPP_CURRENT: "A",
    LPP_FREQUENCY: "Hz",
    LPP_PERCENTAGE: "%",
    LPP_ALTITUDE: "m",
    LPP_CONCENTRATION: "ppm",
    LPP_POWER: "W",
    LPP_DISTANCE: "m",
    LPP_ENERGY: "kWh",
    LPP_DIRECTION: "°",
    LPP_GYROMETER: "°/s",
}


def parse_cayenne_lpp(data: bytes) -> list[dict]:
    """Parse Cayenne LPP encoded telemetry data.

    Returns list of dicts with 'channel', 'type', 'name', 'value', 'unit' keys.
    """
    results = []
    buf = io.BytesIO(data)
    while buf.tell() < len(data) - 1:
        channel = buf.read(1)
        type_byte = buf.read(1)
        if len(channel) < 1 or len(type_byte) < 1:
            break
        ch = channel[0]
        tp = type_byte[0]
        if ch == 0 and tp == 0:
            break
        name = LPP_TYPE_NAMES.get(tp, f"Type {tp}")
        unit = LPP_TYPE_UNITS.get(tp, "")
        try:
            if tp in (LPP_DIGITAL_INPUT, LPP_DIGITAL_OUTPUT, LPP_PRESENCE, LPP_SWITCH):
                val = buf.read(1)[0]
            elif tp in (LPP_ANALOG_INPUT, LPP_ANALOG_OUTPUT):
                val = struct.unpack(">h", buf.read(2))[0] / 100.0
            elif tp == LPP_GENERIC_SENSOR:
                val = struct.unpack(">I", buf.read(4))[0]
            elif tp == LPP_LUMINOSITY:
                val = struct.unpack(">H", buf.read(2))[0]
            elif tp == LPP_TEMPERATURE:
                val = struct.unpack(">h", buf.read(2))[0] / 10.0
            elif tp == LPP_HUMIDITY:
                val = buf.read(1)[0] / 2.0
            elif tp == LPP_PERCENTAGE:
                val = buf.read(1)[0]
            elif tp in (LPP_ACCELEROMETER, LPP_GYROMETER):
                raw = buf.read(6)
                scale = 0.001 if tp == LPP_ACCELEROMETER else 0.01
                val = {
                    "x": struct.unpack(">h", raw[0:2])[0] * scale,
                    "y": struct.unpack(">h", raw[2:4])[0] * scale,
                    "z": struct.unpack(">h", raw[4:6])[0] * scale,
                }
            elif tp == LPP_PRESSURE:
                val = struct.unpack(">H", buf.read(2))[0] / 10.0
            elif tp == LPP_VOLTAGE:
                val = struct.unpack(">h", buf.read(2))[0] / 100.0
            elif tp == LPP_CURRENT:
                val = struct.unpack(">h", buf.read(2))[0] / 1000.0
            elif tp in (LPP_FREQUENCY, LPP_UNIX_TIME):
                val = struct.unpack(">I", buf.read(4))[0]
            elif tp == LPP_ALTITUDE:
                val = struct.unpack(">h", buf.read(2))[0]
            elif tp in (LPP_CONCENTRATION, LPP_POWER, LPP_DIRECTION):
                val = struct.unpack(">H", buf.read(2))[0]
            elif tp in (LPP_DISTANCE, LPP_ENERGY):
                val = struct.unpack(">I", buf.read(4))[0] / 1000.0
            elif tp == LPP_COLOUR:
                raw = buf.read(3)
                val = {"r": raw[0], "g": raw[1], "b": raw[2]}
            elif tp == LPP_GPS:
                raw = buf.read(9)
                lat = int.from_bytes(raw[0:3], "big", signed=True) / 10000.0
                lon = int.from_bytes(raw[3:6], "big", signed=True) / 10000.0
                alt = int.from_bytes(raw[6:9], "big", signed=True) / 100.0
                val = {"lat": lat, "lon": lon, "alt": alt}
            else:
                # Unknown type — can't determine size, stop parsing
                break
            results.append(
                {
                    "channel": ch,
                    "type": tp,
                    "name": name,
                    "value": val,
                    "unit": unit,
                }
            )
        except (struct.error, IndexError):
            break
    return results



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""Message sending, retry, and delivery tracking extracted from MeshyWindow."""

from __future__ import annotations

import logging
import time
import uuid
from datetime import datetime

from gi.repository import GLib
from meshy.models import (
    ChannelMessage,
    Contact,
    Message,
    MessageStatus,
)
from meshy.path_history import PathHistoryService
from meshy.protocol import (
    STATS_TYPE_RADIO,
    build_reset_path,
    build_send_channel_text_msg,
    build_send_text_msg,
    build_set_flood_scope,
)

log = logging.getLogger(__name__)

PATH_RETRIES = 4
FLOOD_RETRIES = 3
MAX_IN_FLIGHT = 6


class MessageController:
    """Owns message send/retry state and delivery tracking."""

    def __init__(self, win):
        self._win = win
        self._path_history: PathHistoryService | None = None
        self.pending_acks: dict[str, str] = {}
        self._acks_by_msg: dict[str, set[str]] = {}
        self.retry_context: dict[str, dict] = {}
        self.sent_msg_queue: list[str] = []
        self.pending_timeouts: dict[str, int] = {}
        self.grace_acks: dict[str, str] = {}
        self._grace_by_msg: dict[str, set[str]] = {}
        self.grace_timeouts: dict[str, int] = {}
        self.pending_flood_attributions: dict[str, list[tuple[int, bool]]] = {}
        self.grace_contexts: dict[str, dict] = {}
        self.last_rx_time: float = 0.0
        self.prev_total_air_secs: int = 0
        self.last_airtime_bump: float = 0.0

    @property
    def path_history(self) -> PathHistoryService:
        if self._path_history is None:
            self._path_history = PathHistoryService(self._win._storage)
        return self._path_history

    @property
    def _device_hash_size(self) -> int:
        return getattr(getattr(self._win, "_device_info", None), "path_hash_mode", 0) + 1

    @staticmethod
    def _determine_strategy(contact: Contact, route_override: str | None) -> str:
        if route_override == "direct":
            return "override_direct"
        if route_override == "flood":
            return "override_flood"
        if contact.force_flood:
            return "force_flood"
        if contact.smart_route_selection:
            return "smart"
        return "traditional"

    def cleanup_on_disconnect(self):
        self._path_history = None
        self.pending_acks.clear()
        self._acks_by_msg.clear()
        self.retry_context.clear()
        for tid in self.pending_timeouts.values():
            GLib.source_remove(tid)
        self.pending_timeouts.clear()
        for tid in self.grace_timeouts.values():
            GLib.source_remove(tid)
        self.grace_timeouts.clear()
        self.grace_acks.clear()
        self._grace_by_msg.clear()
        self.grace_contexts.clear()
        self.pending_flood_attributions.clear()
        self.sent_msg_queue.clear()

    def send_message(self, contact: Contact, text: str, route_override: str | None = None):
        msg_id = str(uuid.uuid4())
        timestamp = int(datetime.now().timestamp())

        message = Message(
            sender_key=contact.public_key,
            text=text,
            timestamp=datetime.fromtimestamp(timestamp),
            is_outgoing=True,
            status=MessageStatus.PENDING,
            message_id=msg_id,
        )
        self._win._chat_view.on_message_sent(message)

        has_path = contact.path_length >= 0 and contact.path_length != 0xFF
        strategy = self._determine_strategy(contact, route_override)
        self.retry_context[msg_id] = {
            "contact": contact,
            "text": text,
            "timestamp": timestamp,
            "attempt": 0,
            "started_with_path": has_path,
            "strategy": strategy,
            "has_path_history": (
                strategy == "smart" and self.path_history.has_paths(contact.public_key_hex)
            ),
            "route_override": route_override,
        }

        def _after_render():
            self._win._storage.save_message(contact.public_key_hex, message)
            contact.last_message_at = message.timestamp
            self._win._frame_handler.schedule_contacts_refresh()
            self._send_message_attempt(msg_id)
            return False

        GLib.idle_add(_after_render)

    def resend_message(self, contact: Contact, message: Message):
        msg_id = message.message_id
        timestamp = int(message.timestamp.timestamp())

        if msg_id in self.pending_timeouts:
            GLib.source_remove(self.pending_timeouts.pop(msg_id))
        for k in self._acks_by_msg.pop(msg_id, set()):
            self.pending_acks.pop(k, None)
        for k in self._grace_by_msg.pop(msg_id, set()):
            self.grace_acks.pop(k, None)
        if msg_id in self.grace_timeouts:
            GLib.source_remove(self.grace_timeouts.pop(msg_id))
        if msg_id in self.sent_msg_queue:
            self.sent_msg_queue.remove(msg_id)

        self._win._storage.update_message_status(msg_id, MessageStatus.PENDING)
        self._win._chat_view.update_message_status(msg_id, MessageStatus.PENDING)

        has_path = contact.path_length >= 0 and contact.path_length != 0xFF
        strategy = self._determine_strategy(contact, None)
        self.retry_context[msg_id] = {
            "contact": contact,
            "text": message.text,
            "timestamp": timestamp,
            "attempt": 0,
            "started_with_path": has_path,
            "strategy": strategy,
            "has_path_history": (
                strategy == "smart" and self.path_history.has_paths(contact.public_key_hex)
            ),
        }
        self._send_message_attempt(msg_id)

    def _send_channel_frame(self, channel_index: int, frame: bytes):
        channel = next((ch for ch in self._win._channels if ch.index == channel_index), None)
        region = channel.region if channel else ""
        if region:
            self._win.send_frame(build_set_flood_scope(region))
            self._win.send_frame(frame)
            GLib.timeout_add(300, self._win._reset_session_scope)
        else:
            self._win.send_frame(frame)

    def send_channel_message(self, channel_index: int, text: str):
        timestamp = int(datetime.now().timestamp())
        self._send_channel_frame(
            channel_index, build_send_channel_text_msg(channel_index, text, timestamp=timestamp)
        )
        name = self._win._self_info.get("name", _("Me"))
        msg = ChannelMessage(
            channel_index=channel_index,
            sender_name=name,
            text=text,
            timestamp=datetime.fromtimestamp(timestamp),
            is_outgoing=True,
        )
        self._win._storage.save_channel_message(msg)
        self._win._channels_view.on_channel_message_sent(msg, timestamp)

    def resend_channel_message(self, channel_index: int, text: str) -> int:
        new_ts = int(datetime.now().timestamp())
        self._send_channel_frame(
            channel_index, build_send_channel_text_msg(channel_index, text, timestamp=new_ts)
        )
        return new_ts

    def delete_message(self, message_id: str):
        self._win._storage.delete_message(message_id)

    def delete_channel_message(self, channel_index: int, timestamp: float, text: str):
        self._win._storage.delete_channel_message(channel_index, timestamp, text)

    def max_retries_for(self, ctx: dict) -> int:
        strategy = ctx.get("strategy", "traditional")
        if strategy == "override_direct":
            return 1
        if strategy in ("override_flood", "force_flood"):
            return FLOOD_RETRIES
        if strategy == "smart":
            return PATH_RETRIES + 1 if ctx.get("has_path_history") else FLOOD_RETRIES
        if ctx.get("started_with_path"):
            return PATH_RETRIES + 1
        return FLOOD_RETRIES

    def send_detail(self, msg_id: str, is_flood: bool) -> str:
        ctx = self.retry_context.get(msg_id)
        contact = ctx.get("contact") if ctx else None

        if is_flood:
            route = _("flood")
        elif contact and contact.path_length == 0:
            route = _("direct")
        else:
            route = _("path")

        if not ctx:
            return _("– {route}").format(route=route)
        attempt = ctx["attempt"] + 1
        max_retries = self.max_retries_for(ctx)
        if attempt <= 1:
            return _("– {route}").format(route=route)
        return _("({attempt}/{total}) – {route}").format(
            attempt=attempt, total=max_retries, route=route
        )

    def _airtime_busy_fraction(self) -> float:
        if self.last_airtime_bump == 0.0:
            return 0.0
        ms = (time.monotonic() - self.last_airtime_bump) * 1000
        if ms >= 8000:
            return 0.0
        return 1.0 - ms / 8000

    def _radio_quiet_delay_ms(self) -> int:
        elapsed = time.monotonic() - self.last_rx_time
        if elapsed > 3.0:
            return 0

        radio = self._win._device_stats.get(STATS_TYPE_RADIO)
        if not radio:
            return 1000

        noise = radio.get("noise_floor", -118)
        noise_severity = max(0.0, min(1.0, (noise + 118) / 30.0))

        snr = radio.get("last_snr", 12.0)
        snr_severity = max(0.0, min(1.0, (12.0 - snr) / 14.0))

        air_busy = self._airtime_busy_fraction()
        severity = max(noise_severity, snr_severity) * 0.82 + air_busy * 0.18

        return int(500 + severity * 7500)

    def _in_flight_count(self) -> int:
        return len(self.sent_msg_queue) + len(self._acks_by_msg)

    def _send_message_attempt(self, msg_id: str):
        ctx = self.retry_context.get(msg_id)
        if not ctx:
            return

        if self._in_flight_count() >= MAX_IN_FLIGHT:
            GLib.timeout_add(500, lambda: self._send_message_attempt(msg_id) or False)
            return

        delay = self._radio_quiet_delay_ms()
        if delay > 0:
            self._win._chat_view.update_message_status(
                msg_id,
                MessageStatus.PENDING,
                _("waiting {:.1f}s (radio busy)").format(delay / 1000),
            )
            GLib.timeout_add(delay, self._do_send_message, msg_id)
        else:
            self._do_send_message(msg_id)

    def _reset_to_flood(self, contact: Contact):
        if contact.path_length >= 0:
            self._win.send_frame(build_reset_path(contact.public_key))
            contact.path_length = -1
            contact.path = b""

    def _restore_original_path(self, ctx: dict):
        orig = ctx.get("original_path")
        if not orig:
            return
        contact = ctx["contact"]
        pl = orig["path_length"]
        if pl < 0:
            self._reset_to_flood(contact)
        else:
            hash_mode = orig["path_hash_size"] - 1
            path_len_byte = (hash_mode << 6) | (pl & 0x3F)
            self._win.set_contact_path(contact, path_len_byte, orig["path"])

    def _apply_traditional_route(self, ctx: dict, contact: Contact):
        if (
            contact.auto_reset_path
            and ctx.get("started_with_path")
            and ctx["attempt"] >= PATH_RETRIES
        ):
            self._reset_to_flood(contact)

    def _apply_smart_route(self, ctx: dict, contact: Contact):
        if not ctx.get("has_path_history") or ctx["attempt"] >= PATH_RETRIES:
            self._reset_to_flood(contact)
            return

        tried = ctx.get("paths_tried", [])
        best = self.path_history.select_path(
            contact.public_key_hex,
            exclude_paths=tried,
        )
        if best is None and tried:
            ctx["paths_tried"] = []
            best = self.path_history.select_path(contact.public_key_hex)
        if best:
            path_bytes_raw = bytes.fromhex(best.path_bytes)
            hash_mode = best.path_hash_size - 1
            path_len_byte = (hash_mode << 6) | (best.hop_count & 0x3F)
            if contact.path.hex() != best.path_bytes or contact.path_length != best.hop_count:
                self._win.set_contact_path(contact, path_len_byte, path_bytes_raw)
        else:
            self._reset_to_flood(contact)

    def _do_send_message(self, msg_id: str) -> bool:
        ctx = self.retry_context.get(msg_id)
        if not ctx:
            return False

        contact_key = ctx["contact"].public_key_hex
        contact = self._win._contacts_by_key.get(contact_key, ctx["contact"])
        ctx["contact"] = contact

        strategy = ctx["strategy"]
        if strategy in ("override_direct", "override_flood", "smart", "traditional"):
            if "original_path" not in ctx:
                ctx["original_path"] = {
                    "path_length": contact.path_length,
                    "path": contact.path,
                    "path_hash_size": contact.path_hash_size,
                }

        if strategy == "override_direct":
            self._win.set_contact_path(contact, 0, b"")
        elif strategy == "override_flood" or strategy == "force_flood":
            self._reset_to_flood(contact)
        elif strategy == "smart":
            try:
                self._apply_smart_route(ctx, contact)
            except Exception:
                log.exception("Smart route selection failed, sending with current path")
        else:
            self._apply_traditional_route(ctx, contact)

        ctx["path_used"] = {
            "path_length": contact.path_length,
            "path_bytes": contact.path.hex(),
            "path_hash_size": contact.path_hash_size,
            "hop_count": max(contact.path_length, 0),
        }
        ctx.setdefault("attempt_paths", {})[ctx["attempt"]] = ctx["path_used"]
        ctx.setdefault("paths_tried", []).append(contact.path.hex())

        self.sent_msg_queue.append(msg_id)
        frame = build_send_text_msg(
            contact.public_key,
            ctx["text"],
            attempt=ctx["attempt"],
            timestamp=ctx["timestamp"],
        )
        self._win.send_frame(frame)
        GLib.timeout_add(10000, self._on_sent_pending_timeout, msg_id)
        return False

    def register_ack_key(self, msg_id: str, ack_key: str):
        ctx = self.retry_context.get(msg_id)
        if ctx:
            ctx.setdefault("ack_paths", {})[ack_key] = ctx.get("attempt_paths", {}).get(
                ctx["attempt"]
            )

    def _on_sent_pending_timeout(self, msg_id: str) -> bool:
        if msg_id in self.sent_msg_queue:
            log.warning(f"Message {msg_id[:8]} stuck in sent queue, forcing retry")
            self.sent_msg_queue.remove(msg_id)
            self._on_message_timeout(msg_id, "")
        return False

    def _on_message_timeout(self, msg_id: str, ack_key: str) -> bool:
        if msg_id in self.pending_timeouts:
            self.pending_timeouts.pop(msg_id)

        ctx = self.retry_context.get(msg_id)
        if ctx:
            self.record_path_result(msg_id, success=False, ack_key=ack_key)
        max_retries = self.max_retries_for(ctx) if ctx else 0
        if ctx and ctx["attempt"] < max_retries - 1:
            ctx["attempt"] += 1
            backoff_ms = 1000 * (1 << (ctx["attempt"] - 1))
            detail = "({current}/{total})".format(current=ctx["attempt"] + 1, total=max_retries)
            self._win._storage.update_message_status(msg_id, MessageStatus.PENDING)
            self._win._chat_view.update_message_status(msg_id, MessageStatus.PENDING, detail)
            GLib.timeout_add(backoff_ms, lambda: self._send_message_attempt(msg_id) or False)
        elif ctx:
            attempt = ctx["attempt"] + 1
            contact = ctx.get("contact")
            started_with_path = ctx.get("started_with_path")
            strategy = ctx.get("strategy", "traditional")
            self._restore_original_path(ctx)
            self.retry_context.pop(msg_id, None)
            self._win._storage.update_message_status(msg_id, MessageStatus.FAILED)
            self._win._chat_view.update_message_status(
                msg_id, MessageStatus.FAILED, f"{attempt}/{max_retries}"
            )
            # Move all ack_keys for this message to grace_acks
            all_ack_keys = list(self._acks_by_msg.pop(msg_id, set()))
            if ack_key and ack_key != "00000000" and ack_key not in all_ack_keys:
                all_ack_keys.append(ack_key)
            for k in all_ack_keys:
                self.pending_acks.pop(k, None)
                self.grace_acks[k] = msg_id
                self._grace_by_msg.setdefault(msg_id, set()).add(k)
            if all_ack_keys:
                grace_id = GLib.timeout_add(30000, self._on_grace_expired, msg_id)
                self.grace_timeouts[msg_id] = grace_id
                if contact:
                    self.grace_contexts[msg_id] = {
                        "contact_key": contact.public_key_hex,
                        "ack_paths": ctx.get("ack_paths", {}),
                        "path_used": ctx.get("path_used"),
                    }
            if strategy == "traditional" and contact:
                contact_key = contact.public_key_hex
                current = self._win._contacts_by_key.get(contact_key)
                if current and current.auto_reset_path and started_with_path:
                    self._reset_to_flood(current)
                    self._win._contacts_view.update_contacts(self._win._contacts)
                    self._win._refresh_chat_header()
            if contact:
                GLib.timeout_add(500, lambda: self._win._frame_handler.send_get_contacts() or False)

        return False

    def record_path_result(
        self, msg_id: str, success: bool, trip_ms: int = 0, ack_key: str | None = None
    ):
        ctx = self.retry_context.get(msg_id)
        if not ctx:
            return
        path_used = None
        if ack_key:
            path_used = ctx.get("ack_paths", {}).get(ack_key)
        if not path_used:
            path_used = ctx.get("path_used")
        if not path_used or path_used["path_length"] < 0:
            # Flood was used - if successful, wait for firmware to set path,
            # then record via refetch callback
            return
        self.path_history.record_result(
            ctx["contact"].public_key_hex,
            path_used["path_bytes"],
            self._device_hash_size,
            path_used["hop_count"],
            success=success,
            trip_ms=trip_ms if success else None,
        )

    GRACE_WEIGHT_INCREMENT = 0.2

    def record_flood_attribution(
        self,
        contact_key: str,
        trip_ms: int = 0,
        late: bool = False,
    ):
        """Credit the contact's firmware-discovered path after a flood success."""
        fresh = self._win._contacts_by_key.get(contact_key)
        if not fresh or fresh.path_length < 0:
            return
        log.debug(
            "Flood attribution%s: crediting path %s for %s (trip=%dms)",
            " (late)" if late else "",
            fresh.path.hex(),
            contact_key[:8],
            trip_ms,
        )
        self.path_history.record_result(
            contact_key,
            fresh.path.hex(),
            self._device_hash_size,
            fresh.path_length,
            success=True,
            trip_ms=trip_ms,
            weight_increment=self.GRACE_WEIGHT_INCREMENT if late else None,
        )

    def _on_grace_expired(self, msg_id: str) -> bool:
        for k in self._grace_by_msg.pop(msg_id, set()):
            self.grace_acks.pop(k, None)
        self.grace_timeouts.pop(msg_id, None)
        self.grace_contexts.pop(msg_id, None)
        return False



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""GeoClue2-based location service for Linux."""

import logging

log = logging.getLogger(__name__)


def get_location_linux(callback):
    """Get current location using GeoClue2.

    Args:
        callback: Function to call with (latitude, longitude, error).
                  If successful, error is None. If failed, lat/lon are None.
    """
    try:
        import gi
        gi.require_version("Geoclue", "2.0")
        from gi.repository import Geoclue

        def _on_ready(source, result, *args):
            try:
                geoclue = Geoclue.Simple.new_finish(result)
                loc = geoclue.get_location()
                lat = loc.get_property("latitude")
                lon = loc.get_property("longitude")
                log.info(f"Location obtained: {lat}, {lon}")
                callback(lat, lon, None)
            except Exception as e:
                log.error(f"Failed to get location: {e}")
                callback(None, None, str(e))

        Geoclue.Simple.new(
            "page.codeberg.sesivany.Meshy",
            Geoclue.AccuracyLevel.EXACT,
            None,
            _on_ready,
        )

    except Exception as e:
        log.error(f"Failed to initialize GeoClue: {e}")
        callback(None, None, str(e))



# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

"""GTK/GIO-based notification service for Linux."""

import logging

from gi.repository import Gio, Gtk

from ..notification_base import NotificationService

log = logging.getLogger(__name__)


class GioNotificationService(NotificationService):
    """Notification service using Gio.Notification (GTK standard on Linux)."""

    def __init__(self, application: Gtk.Application):
        """Initialize notification service.

        Args:
            application: GTK application instance for sending notifications
        """
        self._app = application
        log.debug("GioNotificationService initialized")

    def send_notification(self, title: str, body: str, notification_id: str) -> None:
        """Send a desktop notification using Gio.Notification.

        Args:
            title: Notification title
            body: Notification body text
            notification_id: Unique identifier for this notification
        """
        try:
            notification = Gio.Notification.new(title)
            notification.set_body(body)
            self._app.send_notification(notification_id, notification)
            log.debug(f"Sent notification: {title}")
        except Exception as e:
            log.warning(f"Failed to send notification: {e}")

    def is_supported(self) -> bool:
        """Check if Gio notifications are supported.

        Returns:
            True - Gio.Notification is always available with GTK4
        """
        return True
